# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Networks for continuous control agents."""

import dataclasses
from typing import Optional, Tuple

from dopamine.jax import continuous_networks
from flax import linen as nn
import gin
import jax
from jax import numpy as jnp


@dataclasses.dataclass
class SACEncoderOutputs:
  """The output of a SACEncoder."""
  critic_z: jnp.ndarray
  actor_z: jnp.ndarray


class SACEncoderNetwork(nn.Module):
  """An encoder network for soft actor critic.

  This is the network described in the SAC-AE paper "Improving sample
  efficiency in model-free reinforcement learning from images"
  (Yarats, Zhang, Kostrikov, Amos, Pineau, & Fergus, 2019), and used in
  experiments in the DrQ paper "Image Augmentation Is All You Need:
  Regularizing Deep Reinforcement Learning from Pixels" (Kostrikov, Yarats,
  & Fergus, 2020).
  """

  @nn.compact
  def __call__(self, x):
    # Images are stored in the replay buffer as uint8.
    x = x.astype(jnp.float32) / 255.0

    # Flatten the last dimension (normally to deal with stacked rgb frames)
    if len(x.shape) > 3:
      x = x.reshape((*x.shape[:2], -1))

    kernel_init = nn.initializers.orthogonal()
    x = nn.Conv(
        features=32,
        kernel_size=(3, 3),
        strides=(2, 2),
        kernel_init=kernel_init)(x)
    x = nn.relu(x)
    x = nn.Conv(
        features=32,
        kernel_size=(3, 3),
        strides=(1, 1),
        kernel_init=kernel_init)(x)
    x = nn.relu(x)
    x = nn.Conv(
        features=32,
        kernel_size=(3, 3),
        strides=(1, 1),
        kernel_init=kernel_init)(x)
    x = nn.relu(x)
    x = nn.Conv(
        features=32,
        kernel_size=(3, 3),
        strides=(1, 1),
        kernel_init=kernel_init)(x)
    x = nn.relu(x)
    x = jnp.reshape(x, -1)  # Flatten

    critic_z = nn.Dense(features=50, kernel_init=kernel_init)(x)
    critic_z = nn.LayerNorm()(critic_z)
    critic_z = nn.tanh(critic_z)

    # Only the critic should train the convolution layers, so stop the
    # gradients from the actor.
    actor_z = nn.Dense(
        features=50, kernel_init=kernel_init)(
            jax.lax.stop_gradient(x))
    actor_z = nn.LayerNorm()(actor_z)
    actor_z = nn.tanh(actor_z)

    return SACEncoderOutputs(critic_z, actor_z)


@gin.configurable
class SACConvNetwork(nn.Module):
  """A convolutional value and policy networks for SAC.

  This uses the SAC-AE network for processing images. For more information,
  view SACEncoderNetwork.
  """
  action_shape: Tuple[int, ...]
  num_layers: int = 2
  hidden_units: int = 256
  action_limits: Optional[Tuple[Tuple[float, ...], Tuple[float, ...]]] = None

  def setup(self):
    self._encoder = SACEncoderNetwork()
    self._sac_network = continuous_networks.SACNetwork(
        self.action_shape, self.num_layers,
        self.hidden_units, self.action_limits)

  def __call__(self,
               state: jnp.ndarray,
               key: jnp.ndarray,
               mean_action: bool = True) -> continuous_networks.SacOutput:
    """Calls the SAC actor/critic networks.

    This has two important purposes:
      1. It is used to initialize all parameters of both networks.
      2. It is used to efficiently calculate the outputs of both the actor
        and critic networks on a single input.

    Args:
      state: An input state.
      key: A PRNGKey to use to sample an action from the actor's output
        distribution.
      mean_action: If True, it will use the actor's mean action to feed to the
        value network. Otherwise, it will use the sampled action.

    Returns:
      A named tuple containing the outputs from both networks.
    """
    encoding = self._encoder(state)

    actor_output = self._sac_network.actor(encoding.actor_z, key)
    action = actor_output.mean_action if mean_action else actor_output.sampled_action
    critic_output = self._sac_network.critic(encoding.critic_z, action)

    return continuous_networks.SacOutput(actor_output, critic_output)

  def actor(self, state: jnp.ndarray,
            key: jnp.ndarray) -> continuous_networks.SacActorOutput:
    """Calls the SAC actor network.

    This can be called using network_def.apply(..., method=network_def.actor).

    Args:
      state: An input state.
      key: A PRNGKey to use to sample an action from the actor's output
        distribution.

    Returns:
      A named tuple containing a sampled action, the mean action, and the
        likelihood of the sampled action.
    """
    encoding = self._encoder(state)
    return self._sac_network.actor(encoding.actor_z, key)

  def critic(self, state: jnp.ndarray,
             action: jnp.ndarray) -> continuous_networks.SacCriticOutput:
    """Calls the SAC critic network.

    SAC uses two Q networks to reduce overestimation bias.
    This can be called using network_def.apply(..., method=network_def.critic).

    Args:
      state: An input state.
      action: An action to compute the value of.

    Returns:
      A named tuple containing the Q values of each network.
    """
    encoding = self._encoder(state)
    return self._sac_network.critic(encoding.critic_z, action)



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Various losses used by the Dopamine JAX agents."""
from flax import linen as nn
import jax.numpy as jnp


def huber_loss(targets: jnp.array,
               predictions: jnp.array,
               delta: float = 1.0) -> jnp.ndarray:
  """Implementation of the Huber loss with threshold delta.

  Let `x = |targets - predictions|`, the Huber loss is defined as:
  `0.5 * x^2` if `x <= delta`
  `0.5 * delta^2 + delta * (x - delta)` otherwise.

  Args:
    targets: Target values.
    predictions: Prediction values.
    delta: Threshold.

  Returns:
    Huber loss.
  """
  x = jnp.abs(targets - predictions)
  return jnp.where(x <= delta,
                   0.5 * x**2,
                   0.5 * delta**2 + delta * (x - delta))


def mse_loss(targets: jnp.array, predictions: jnp.array) -> jnp.ndarray:
  """Implementation of the mean squared error loss."""
  return jnp.power((targets - predictions), 2)


def softmax_cross_entropy_loss_with_logits(labels: jnp.array,
                                           logits: jnp.array) -> jnp.ndarray:
  """Implementation of the softmax cross entropy loss."""
  return -jnp.sum(labels * nn.log_softmax(logits))



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compact implementation of a Soft Actor-Critic agent in JAX.

Based on agent described in
  "Soft Actor-Critic Algorithms and Applications"
  by Tuomas Haarnoja et al.
  https://arxiv.org/abs/1812.05905
"""

import functools
import math
import operator
import time
from typing import Any, Mapping, Tuple

from absl import logging
from dopamine.jax import continuous_networks
from dopamine.jax import losses
from dopamine.jax.agents.dqn import dqn_agent
# pylint: disable=unused-import
# This enables (experimental) networks for SAC from pixels.
# Note, that the full name import is required to avoid a naming
# collision with the short name import (continuous_networks) above.
import dopamine.labs.sac_from_pixels.continuous_networks
# pylint: enable=unused-import
from dopamine.replay_memory import circular_replay_buffer
import flax
from flax import linen as nn
import gin
import jax
import jax.numpy as jnp
import numpy as onp
import optax
import tensorflow as tf


try:
  logging.warning(
      ('Setting tf to CPU only, to avoid OOM. '
       'See https://jax.readthedocs.io/en/latest/gpu_memory_allocation.html '
       'for more information.'))
  tf.config.set_visible_devices([], 'GPU')
except tf.errors.NotFoundError:
  logging.info(
      ('Unable to modify visible devices. '
       'If you don\'t have a GPU, this is expected.'))


gin.constant('sac_agent.IMAGE_DTYPE', onp.uint8)
gin.constant('sac_agent.STATE_DTYPE', onp.float32)


@functools.partial(jax.jit, static_argnums=(0, 1, 2))
def train(network_def: nn.Module,
          optim: optax.GradientTransformation,
          alpha_optim: optax.GradientTransformation,
          optimizer_state: jnp.ndarray,
          alpha_optimizer_state: jnp.ndarray,
          network_params: flax.core.FrozenDict,
          target_params: flax.core.FrozenDict,
          log_alpha: jnp.ndarray,
          key: jnp.ndarray,
          states: jnp.ndarray,
          actions: jnp.ndarray,
          next_states: jnp.ndarray,
          rewards: jnp.ndarray,
          terminals: jnp.ndarray,
          cumulative_gamma: float,
          target_entropy: float,
          reward_scale_factor: float) -> Mapping[str, Any]:
  """Run the training step.

  Returns a list of updated values and losses.

  Args:
    network_def: The SAC network definition.
    optim: The SAC optimizer (which also wraps the SAC parameters).
    alpha_optim: The optimizer for alpha.
    optimizer_state: The SAC optimizer state.
    alpha_optimizer_state: The alpha optimizer state.
    network_params: Parameters for SAC's online network.
    target_params: The parameters for SAC's target network.
    log_alpha: Parameters for alpha network.
    key: An rng key to use for random action selection.
    states: A batch of states.
    actions: A batch of actions.
    next_states: A batch of next states.
    rewards: A batch of rewards.
    terminals: A batch of terminals.
    cumulative_gamma: The discount factor to use.
    target_entropy: The target entropy for the agent.
    reward_scale_factor: A factor by which to scale rewards.

  Returns:
    A mapping from string keys to values, including updated optimizers and
      training statistics.
  """
  # Get the models from all the optimizers.
  frozen_params = network_params  # For use in loss_fn without apply gradients

  batch_size = states.shape[0]
  actions = jnp.reshape(actions, (batch_size, -1))  # Flatten

  def loss_fn(
      params: flax.core.FrozenDict, log_alpha: flax.core.FrozenDict,
      state: jnp.ndarray, action: jnp.ndarray, reward: jnp.ndarray,
      next_state: jnp.ndarray, terminal: jnp.ndarray,
      rng: jnp.ndarray) -> Tuple[jnp.ndarray, Mapping[str, jnp.ndarray]]:
    """Calculates the loss for one transition.

    Args:
      params: Parameters for the SAC network.
      log_alpha: SAC's log_alpha parameter.
      state: A single state vector.
      action: A single action vector.
      reward: A reward scalar.
      next_state: A next state vector.
      terminal: A terminal scalar.
      rng: An RNG key to use for sampling actions.

    Returns:
      A tuple containing 1) the combined SAC loss and 2) a mapping containing
        statistics from the loss step.
    """
    rng1, rng2 = jax.random.split(rng, 2)

    # J_Q(\theta) from equation (5) in paper.
    q_value_1, q_value_2 = network_def.apply(
        params, state, action, method=network_def.critic)
    q_value_1 = jnp.squeeze(q_value_1)
    q_value_2 = jnp.squeeze(q_value_2)

    target_outputs = network_def.apply(target_params, next_state, rng1, True)
    target_q_value_1, target_q_value_2 = target_outputs.critic
    target_q_value = jnp.squeeze(
        jnp.minimum(target_q_value_1, target_q_value_2))

    alpha_value = jnp.exp(log_alpha)
    log_prob = target_outputs.actor.log_probability
    target = reward_scale_factor * reward + cumulative_gamma * (
        target_q_value - alpha_value * log_prob) * (1. - terminal)
    target = jax.lax.stop_gradient(target)
    critic_loss_1 = losses.mse_loss(q_value_1, target)
    critic_loss_2 = losses.mse_loss(q_value_2, target)
    critic_loss = jnp.mean(critic_loss_1 + critic_loss_2)

    # J_{\pi}(\phi) from equation (9) in paper.
    mean_action, sampled_action, action_log_prob = network_def.apply(
        params, state, rng2, method=network_def.actor)

    # We use frozen_params so that gradients can flow back to the actor without
    # being used to update the critic.
    q_value_no_grad_1, q_value_no_grad_2 = network_def.apply(
        frozen_params, state, sampled_action, method=network_def.critic)
    no_grad_q_value = jnp.squeeze(
        jnp.minimum(q_value_no_grad_1, q_value_no_grad_2))
    alpha_value = jnp.exp(jax.lax.stop_gradient(log_alpha))
    policy_loss = jnp.mean(alpha_value * action_log_prob - no_grad_q_value)

    # J(\alpha) from equation (18) in paper.
    entropy_diff = -action_log_prob - target_entropy
    alpha_loss = jnp.mean(log_alpha * jax.lax.stop_gradient(entropy_diff))

    # Giving a smaller weight to the critic empirically gives better results
    combined_loss = 0.5 * critic_loss + 1.0 * policy_loss + 1.0 * alpha_loss
    return combined_loss, {
        'critic_loss': critic_loss,
        'policy_loss': policy_loss,
        'alpha_loss': alpha_loss,
        'critic_value_1': q_value_1,
        'critic_value_2': q_value_2,
        'target_value_1': target_q_value_1,
        'target_value_2': target_q_value_2,
        'mean_action': mean_action
    }

  grad_fn = jax.vmap(
      jax.value_and_grad(loss_fn, argnums=(0, 1), has_aux=True),
      in_axes=(None, None, 0, 0, 0, 0, 0, 0))

  rng = jnp.stack(jax.random.split(key, num=batch_size))
  (_, aux_vars), gradients = grad_fn(network_params, log_alpha, states, actions,
                                     rewards, next_states, terminals, rng)

  # This calculates the mean gradient/aux_vars using the individual
  # gradients/aux_vars from each item in the batch.
  gradients = jax.tree_map(functools.partial(jnp.mean, axis=0), gradients)
  aux_vars = jax.tree_map(functools.partial(jnp.mean, axis=0), aux_vars)
  network_gradient, alpha_gradient = gradients

  # Apply gradients to all the optimizers.
  updates, optimizer_state = optim.update(network_gradient, optimizer_state,
                                          params=network_params)
  network_params = optax.apply_updates(network_params, updates)
  alpha_updates, alpha_optimizer_state = alpha_optim.update(
      alpha_gradient, alpha_optimizer_state, params=log_alpha)
  log_alpha = optax.apply_updates(log_alpha, alpha_updates)

  # Compile everything in a dict.
  returns = {
      'network_params': network_params,
      'log_alpha': log_alpha,
      'optimizer_state': optimizer_state,
      'alpha_optimizer_state': alpha_optimizer_state,
      'Losses/Critic': aux_vars['critic_loss'],
      'Losses/Actor': aux_vars['policy_loss'],
      'Losses/Alpha': aux_vars['alpha_loss'],
      'Values/CriticValues1': jnp.mean(aux_vars['critic_value_1']),
      'Values/CriticValues2': jnp.mean(aux_vars['critic_value_2']),
      'Values/TargetValues1': jnp.mean(aux_vars['target_value_1']),
      'Values/TargetValues2': jnp.mean(aux_vars['target_value_2']),
      'Values/Alpha': jnp.exp(log_alpha),
  }
  for i, a in enumerate(aux_vars['mean_action']):
    returns.update({f'Values/MeanActions{i}': a})
  return returns


@functools.partial(jax.jit, static_argnums=0)
def select_action(network_def, params, state, rng, eval_mode=False):
  """Sample an action to take from the current policy network.

  This obtains a mean and variance from the input policy network, and samples an
  action using a Gaussian distribution.

  Args:
    network_def: Linen Module to use for inference.
    params: Linen params (frozen dict) to use for inference.
    state: input state to use for inference.
    rng: Jax random number generator.
    eval_mode: bool, whether in eval mode.

  Returns:
    rng: Jax random number generator.
    action: int, the selected action.
  """
  rng, rng2 = jax.random.split(rng)
  greedy_a, sampled_a, _ = network_def.apply(
      params, state, rng2, method=network_def.actor)
  return rng, jnp.where(eval_mode, greedy_a, sampled_a)


@gin.configurable
class SACAgent(dqn_agent.JaxDQNAgent):
  """A JAX implementation of the SAC agent."""

  def __init__(self,
               action_shape,
               action_limits,
               observation_shape,
               action_dtype=jnp.float32,
               observation_dtype=jnp.float32,
               reward_scale_factor=1.0,
               stack_size=1,
               network=continuous_networks.SACNetwork,
               num_layers=2,
               hidden_units=256,
               gamma=0.99,
               update_horizon=1,
               min_replay_history=20000,
               update_period=1,
               target_update_type='soft',
               target_update_period=1000,
               target_smoothing_coefficient=0.005,
               target_entropy=None,
               eval_mode=False,
               optimizer='adam',
               summary_writer=None,
               summary_writing_frequency=500,
               allow_partial_reload=False,
               seed=None):
    r"""Initializes the agent and constructs the necessary components.

    Args:
      action_shape: int or tuple, dimensionality of the action space.
      action_limits: pair of lower and higher bounds for actions.
      observation_shape: tuple of ints describing the observation shape.
      action_dtype: jnp.dtype, specifies the type of the actions.
      observation_dtype: jnp.dtype, specifies the type of the observations.
      reward_scale_factor: float, factor by which to scale rewards.
      stack_size: int, number of frames to use in state stack.
      network: Jax network to use for training.
      num_layers: int, number of layers in the network.
      hidden_units: int, number of hidden units in the network.
      gamma: float, discount factor with the usual RL meaning.
      update_horizon: int, horizon at which updates are performed, the 'n' in
        n-step update.
      min_replay_history: int, number of transitions that should be experienced
        before the agent begins training its value function.
      update_period: int, period between DQN updates.
      target_update_type: str, if 'hard', will perform a hard update of the
        target network every target_update_period training steps; if 'soft',
        will use target_smoothing_coefficient to update the target network at
        every training step.
      target_update_period: int, frequency with which to update target network
        when in 'hard' mode.
      target_smoothing_coefficient: float, smoothing coefficient for target
        network updates (\tau in paper) when in 'soft' mode.
      target_entropy: float or None, the target entropy for training alpha. If
        None, it will default to the half the negative of the number of action
        dimensions.
      eval_mode: bool, True for evaluation and False for training.
      optimizer: str, name of optimizer to use.
      summary_writer: SummaryWriter object for outputting training statistics.
      summary_writing_frequency: int, frequency with which summaries will be
        written. Lower values will result in slower training.
      allow_partial_reload: bool, whether we allow reloading a partial agent
        (for instance, only the network parameters).
      seed: int, a seed for SAC's internal RNG, used for initialization and
        sampling actions.
    """
    assert isinstance(observation_shape, tuple)
    # If we're performing hard updates, we force the smoothing coefficient to 1.
    if target_update_type == 'hard':
      target_smoothing_coefficient = 1.0

    if isinstance(action_shape, int):
      action_shape = (action_shape,)

    # If target_entropy is None, set to default value.
    if target_entropy is None:
      action_dim = functools.reduce(operator.mul, action_shape, 1.0)
      target_entropy = -0.5 * action_dim
    seed = int(time.time() * 1e6) if seed is None else seed
    logging.info('Creating %s agent with the following parameters:',
                 self.__class__.__name__)
    logging.info('\t action_shape: %s', action_shape)
    logging.info('\t action_dtype: %s', action_dtype)
    logging.info('\t action_limits: %s', action_limits)
    logging.info('\t observation_shape: %s', observation_shape)
    logging.info('\t observation_dtype: %s', observation_dtype)
    logging.info('\t reward_scale_factor: %f', reward_scale_factor)
    logging.info('\t num_layers: %d', num_layers)
    logging.info('\t hidden_units: %d', hidden_units)
    logging.info('\t gamma: %f', gamma)
    logging.info('\t update_horizon: %f', update_horizon)
    logging.info('\t min_replay_history: %d', min_replay_history)
    logging.info('\t update_period: %d', update_period)
    logging.info('\t target_update_type: %s', target_update_type)
    logging.info('\t target_update_period: %d', target_update_period)
    logging.info('\t target_smoothing_coefficient: %f',
                 target_smoothing_coefficient)
    logging.info('\t target_entropy: %f', target_entropy)
    logging.info('\t optimizer: %s', optimizer)
    logging.info('\t seed: %d', seed)

    self.action_shape = action_shape
    self.action_dtype = action_dtype
    self.observation_shape = tuple(observation_shape)
    self.observation_dtype = observation_dtype
    self.reward_scale_factor = reward_scale_factor
    self.stack_size = stack_size
    self.action_limits = action_limits
    action_limits = tuple(tuple(x.reshape(-1)) for x in action_limits)
    self.network_def = network(action_shape, num_layers, hidden_units,
                               action_limits)
    self.gamma = gamma
    self.update_horizon = update_horizon
    self.cumulative_gamma = math.pow(gamma, update_horizon)
    self.min_replay_history = min_replay_history
    self.update_period = update_period
    self.target_update_type = target_update_type
    self.target_update_period = target_update_period
    self.target_smoothing_coefficient = target_smoothing_coefficient
    self.target_entropy = target_entropy
    self.eval_mode = eval_mode
    self.training_steps = 0
    self.summary_writer = summary_writer
    self.summary_writing_frequency = summary_writing_frequency
    self.allow_partial_reload = allow_partial_reload

    self._rng = jax.random.PRNGKey(seed)
    state_shape = self.observation_shape + (stack_size,)
    self.state = onp.zeros(state_shape)
    self._replay = self._build_replay_buffer()
    self._optimizer_name = optimizer
    self._build_networks_and_optimizer()

    # Variables to be initialized by the agent once it interacts with the
    # environment.
    self._observation = None
    self._last_observation = None

  def _build_networks_and_optimizer(self):
    self._rng, init_key = jax.random.split(self._rng)

    # We can reuse init_key safely for the action selection key
    # since it is only used for shape inference during initialization.
    self.network_params = self.network_def.init(init_key, self.state, init_key)
    self.network_optimizer = dqn_agent.create_optimizer(self._optimizer_name)
    self.optimizer_state = self.network_optimizer.init(self.network_params)

    # TODO(joshgreaves): Find a way to just copy the critic params
    self.target_params = self.network_params

    # \alpha network
    self.log_alpha = jnp.zeros(1)
    self.alpha_optimizer = dqn_agent.create_optimizer(self._optimizer_name)
    self.alpha_optimizer_state = self.alpha_optimizer.init(self.log_alpha)

  def _build_replay_buffer(self):
    """Creates the replay buffer used by the agent."""
    return circular_replay_buffer.OutOfGraphReplayBuffer(
        observation_shape=self.observation_shape,
        stack_size=self.stack_size,
        update_horizon=self.update_horizon,
        gamma=self.gamma,
        observation_dtype=self.observation_dtype,
        action_shape=self.action_shape,
        action_dtype=self.action_dtype)

  def _maybe_sync_weights(self):
    """Syncs the target weights with the online weights."""
    if (self.target_update_type == 'hard' and
        self.training_steps % self.target_update_period != 0):
      return

    def _sync_weights(target_p, online_p):
      return (self.target_smoothing_coefficient * online_p +
              (1 - self.target_smoothing_coefficient) * target_p)

    self.target_params = jax.tree_multimap(_sync_weights, self.target_params,
                                           self.network_params)

  def begin_episode(self, observation):
    """Returns the agent's first action for this episode.

    Args:
      observation: numpy array, the environment's initial observation.

    Returns:
      np.ndarray, the selected action.
    """
    self._reset_state()
    self._record_observation(observation)

    if not self.eval_mode:
      self._train_step()

    if self._replay.add_count > self.min_replay_history:
      self._rng, self.action = select_action(self.network_def,
                                             self.network_params, self.state,
                                             self._rng, self.eval_mode)
    else:
      self._rng, action_rng = jax.random.split(self._rng)
      self.action = jax.random.uniform(action_rng, self.action_shape,
                                       self.action_dtype, self.action_limits[0],
                                       self.action_limits[1])
    self.action = onp.asarray(self.action)
    return self.action

  def step(self, reward, observation):
    """Records the most recent transition and returns the agent's next action.

    We store the observation of the last time step since we want to store it
    with the reward.

    Args:
      reward: float, the reward received from the agent's most recent action.
      observation: numpy array, the most recent observation.

    Returns:
      int, the selected action.
    """
    self._last_observation = self._observation
    self._record_observation(observation)

    if not self.eval_mode:
      self._store_transition(self._last_observation, self.action, reward, False)
      self._train_step()

    if self._replay.add_count > self.min_replay_history:
      self._rng, self.action = select_action(self.network_def,
                                             self.network_params, self.state,
                                             self._rng, self.eval_mode)
    else:
      self._rng, action_rng = jax.random.split(self._rng)
      self.action = jax.random.uniform(action_rng, self.action_shape,
                                       self.action_dtype, self.action_limits[0],
                                       self.action_limits[1])
    self.action = onp.asarray(self.action)
    return self.action

  def _train_step(self):
    """Runs a single training step.

    Runs training if both:
      (1) A minimum number of frames have been added to the replay buffer.
      (2) `training_steps` is a multiple of `update_period`.

    Also, syncs weights from online_network to target_network if training steps
    is a multiple of target update period.
    """
    if self._replay.add_count > self.min_replay_history:
      if self.training_steps % self.update_period == 0:
        self._sample_from_replay_buffer()
        self._rng, key = jax.random.split(self._rng)

        train_returns = train(
            self.network_def, self.network_optimizer, self.alpha_optimizer,
            self.optimizer_state, self.alpha_optimizer_state,
            self.network_params, self.target_params, self.log_alpha,
            key, self.replay_elements['state'],
            self.replay_elements['action'], self.replay_elements['next_state'],
            self.replay_elements['reward'], self.replay_elements['terminal'],
            self.cumulative_gamma, self.target_entropy,
            self.reward_scale_factor)

        self.network_params = train_returns['network_params']
        self.optimizer_state = train_returns['optimizer_state']
        self.log_alpha = train_returns['log_alpha']
        self.alpha_optimizer_state = train_returns['alpha_optimizer_state']

        if (self.summary_writer is not None and
            self.training_steps > 0 and
            self.training_steps % self.summary_writing_frequency == 0):

          for k in train_returns:
            if k.startswith('Losses') or k.startswith('Values'):
              self.summary_writer.scalar(k, train_returns[k],
                                         self.training_steps)
          self.summary_writer.flush()
        self._maybe_sync_weights()
    self.training_steps += 1

  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    """Returns a self-contained bundle of the agent's state.

    This is used for checkpointing. It will return a dictionary containing all
    non-TensorFlow objects (to be saved into a file by the caller), and it saves
    all TensorFlow objects into a checkpoint file.

    Args:
      checkpoint_dir: str, directory where TensorFlow objects will be saved.
      iteration_number: int, iteration number to use for naming the checkpoint
        file.

    Returns:
      A dict containing additional Python objects to be checkpointed by the
        experiment. If the checkpoint directory does not exist, returns None.
    """
    if not tf.io.gfile.exists(checkpoint_dir):
      return None
    # Checkpoint the out-of-graph replay buffer.
    self._replay.save(checkpoint_dir, iteration_number)
    bundle_dictionary = {
        'state': self.state,
        'training_steps': self.training_steps,
        'network_params': self.network_params,
        'optimizer_state': self.optimizer_state,
        'target_params': self.target_params,
        'log_alpha': self.log_alpha,
        'alpha_optimizer_state': self.alpha_optimizer_state,
    }
    return bundle_dictionary

  def unbundle(self, checkpoint_dir, iteration_number, bundle_dictionary):
    """Restores the agent from a checkpoint.

    Restores the agent's Python objects to those specified in bundle_dictionary,
    and restores the TensorFlow objects to those specified in the
    checkpoint_dir. If the checkpoint_dir does not exist, will not reset the
      agent's state.

    Args:
      checkpoint_dir: str, path to the checkpoint saved.
      iteration_number: int, checkpoint version, used when restoring the replay
        buffer.
      bundle_dictionary: dict, containing additional Python objects owned by the
        agent.

    Returns:
      bool, True if unbundling was successful.
    """
    try:
      # self._replay.load() will throw a NotFoundError if it does not find all
      # the necessary files.
      self._replay.load(checkpoint_dir, iteration_number)
    except tf.errors.NotFoundError:
      if not self.allow_partial_reload:
        # If we don't allow partial reloads, we will return False.
        return False
      logging.warning('Unable to reload replay buffer!')
    if bundle_dictionary is not None:
      self.state = bundle_dictionary['state']
      self.training_steps = bundle_dictionary['training_steps']

      self.network_params = bundle_dictionary['network_params']
      self.network_optimizer = dqn_agent.create_optimizer(self._optimizer_name)
      self.optimizer_state = bundle_dictionary['optimizer_state']
      self.target_params = bundle_dictionary['target_params']
      self.log_alpha = bundle_dictionary['log_alpha']
      self.alpha_optimizer = dqn_agent.create_optimizer(self._optimizer_name)
      self.alpha_optimizer_state = bundle_dictionary['alpha_optimizer_state']
    elif not self.allow_partial_reload:
      return False
    else:
      logging.warning("Unable to reload the agent's parameters!")
    return True



# coding=utf-8
# Copyright 2021 The Tandem DQN authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tandem DQN agent (one active, one passive agent)."""

import functools

from absl import logging
from dopamine.jax import losses
from dopamine.jax.agents.dqn import dqn_agent
import gin
import jax
import jax.numpy as jnp
import numpy as onp
import optax
import tensorflow as tf


@functools.partial(jax.jit, static_argnums=(0, 3, 10, 11, 12))
def train(network_def, online_params, target_params, optimizer, optimizer_state,
          states, actions, next_states, rewards, terminals, cumulative_gamma,
          loss_type='huber', double_dqn=True):
  """Run the training step."""
  def loss_fn(params, target):
    def q_online(state):
      return network_def.apply(params, state)

    q_values = jax.vmap(q_online)(states).q_values
    q_values = jnp.squeeze(q_values)
    replay_chosen_q = jax.vmap(lambda x, y: x[y])(q_values, actions)
    if loss_type == 'huber':
      loss = jnp.mean(jax.vmap(losses.huber_loss)(target, replay_chosen_q))
    else:
      loss = jnp.mean(jax.vmap(losses.mse_loss)(target, replay_chosen_q))
    return loss

  def q_online(state):
    return network_def.apply(online_params, state)

  def q_target(state):
    return network_def.apply(target_params, state)

  target = target_q(q_online,
                    q_target,
                    next_states,
                    rewards,
                    terminals,
                    cumulative_gamma,
                    double_dqn)
  grad_fn = jax.value_and_grad(loss_fn)
  loss, grad = grad_fn(online_params, target)
  updates, optimizer_state = optimizer.update(grad, optimizer_state)
  online_params = optax.apply_updates(online_params, updates)
  return optimizer_state, online_params, loss


def target_q(online_network, target_network, next_states, rewards, terminals,
             cumulative_gamma, double_dqn):
  """Compute the target Q-value."""
  if double_dqn:
    next_state_q_vals_for_argmax = jax.vmap(
        online_network, in_axes=(0))(next_states).q_values
  else:
    next_state_q_vals_for_argmax = jax.vmap(
        target_network, in_axes=(0))(next_states).q_values
  next_state_q_vals_for_argmax = jnp.squeeze(next_state_q_vals_for_argmax)
  next_argmax = jnp.argmax(next_state_q_vals_for_argmax, axis=1)
  q_values = jax.vmap(
      target_network, in_axes=(0))(next_states).q_values
  replay_next_qt_max = jax.vmap(lambda t, u: t[u])(q_values, next_argmax)
  return jax.lax.stop_gradient(rewards + cumulative_gamma * replay_next_qt_max *
                               (1. - terminals))


@gin.configurable
class TandemDQNAgent(dqn_agent.JaxDQNAgent):
  """Tandem DQN agent."""

  def __init__(self, num_actions, double_dqn=True, summary_writer=None):
    self._double_dqn = double_dqn
    super().__init__(num_actions, summary_writer=summary_writer)

  def _build_networks_and_optimizer(self):
    self._rng, active_rng, passive_rng = jax.random.split(self._rng, 3)
    # Initialize active networks.
    self.active_online_params = self.network_def.init(active_rng,
                                                      x=self.state)
    self.active_optimizer = dqn_agent.create_optimizer(self._optimizer_name)
    self.active_optimizer_state = self.active_optimizer.init(
        self.active_online_params)
    self.active_target_params = self.active_online_params
    # Initialize passive network with the regular network.
    self.passive_online_params = self.network_def.init(passive_rng,
                                                       x=self.state)
    self.passive_optimizer = dqn_agent.create_optimizer(self._optimizer_name)
    self.passive_optimizer_state = self.passive_optimizer.init(
        self.passive_online_params)
    self.passive_target_params = self.passive_online_params

  def _sync_weights(self):
    """Syncs the target_params with the online_params."""
    self.active_target_params = self.active_online_params
    self.passive_target_params = self.passive_online_params

  def _select_action(self, params):
    self._rng, action = dqn_agent.select_action(self.network_def,
                                                params,
                                                self.state,
                                                self._rng,
                                                self.num_actions,
                                                self.eval_mode,
                                                self.epsilon_eval,
                                                self.epsilon_train,
                                                self.epsilon_decay_period,
                                                self.training_steps,
                                                self.min_replay_history,
                                                self.epsilon_fn)
    action = onp.asarray(action)
    return action

  def begin_episode(self, agent_type, observation):
    self._reset_state()
    self._record_observation(observation)
    if agent_type == 'passive':
      params = self.passive_online_params
    else:
      params = self.active_online_params
      if not self.eval_mode:
        self._train_step()

    action = self._select_action(params)
    if agent_type == 'active':
      self.action = action
    return action

  def step(self, agent_type, reward, observation):
    self._last_observation = self._observation
    self._record_observation(observation)
    if agent_type == 'passive':
      params = self.passive_online_params
    else:
      params = self.active_online_params
      if not self.eval_mode:
        self._store_transition(
            self._last_observation, self.action, reward, False)
        self._train_step()

    action = self._select_action(params)
    if agent_type == 'active':
      self.action = onp.asarray(action)
    return action

  def _train_step(self):
    """Runs a single tandem training step.

    Runs training if both:
      (1) A minimum number of frames have been added to the replay buffer.
      (2) `training_steps` is a multiple of `update_period`.

    Also, syncs weights from online to target if training steps is a multiple of
    target update period.
    """
    # Run a train op at the rate of self.update_period if enough training steps
    # have been run. This matches the Nature DQN behaviour.
    if self._replay.add_count > self.min_replay_history:
      if self.training_steps % self.update_period == 0:
        self._sample_from_replay_buffer()
        self.active_optimizer_state, self.active_online_params, active_loss = (
            dqn_agent.train(
                self.network_def,
                self.active_online_params,
                self.active_target_params,
                self.active_optimizer,
                self.active_optimizer_state,
                self.replay_elements['state'],
                self.replay_elements['action'],
                self.replay_elements['next_state'],
                self.replay_elements['reward'],
                self.replay_elements['terminal'],
                self.cumulative_gamma,
                self._double_dqn))
        (self.passive_optimizer_state, self.passive_online_params,
         passive_loss) = dqn_agent.train(
             self.network_def,
             self.passive_online_params,
             self.passive_target_params,
             self.passive_optimizer,
             self.passive_optimizer_state,
             self.replay_elements['state'],
             self.replay_elements['action'],
             self.replay_elements['next_state'],
             self.replay_elements['reward'],
             self.replay_elements['terminal'],
             self.cumulative_gamma,
             self._double_dqn)
        if (self.summary_writer is not None and
            self.training_steps > 0 and
            self.training_steps % self.summary_writing_frequency == 0):
          values = [tf.compat.v1.Summary.Value(tag='Losses/Active',
                                               simple_value=active_loss),
                    tf.compat.v1.Summary.Value(tag='Losses/Passive',
                                               simple_value=passive_loss)]
          summary = tf.compat.v1.Summary(value=values)
          self.summary_writer.add_summary(summary, self.training_steps)
          self.summary_writer.flush()
      if self.training_steps % self.target_update_period == 0:
        self._sync_weights()

    self.training_steps += 1

  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    """Returns a self-contained bundle of the agent's state.

    This is used for checkpointing. It will return a dictionary containing all
    non-TensorFlow objects (to be saved into a file by the caller), and it saves
    all TensorFlow objects into a checkpoint file.

    Args:
      checkpoint_dir: str, directory where TensorFlow objects will be saved.
      iteration_number: int, iteration number to use for naming the checkpoint
        file.

    Returns:
      A dict containing additional Python objects to be checkpointed by the
        experiment. If the checkpoint directory does not exist, returns None.
    """
    if not tf.io.gfile.exists(checkpoint_dir):
      return None
    # Checkpoint the out-of-graph replay buffer.
    self._replay.save(checkpoint_dir, iteration_number)
    bundle_dictionary = {
        'state': self.state,
        'training_steps': self.training_steps,
        'active_online_params': self.active_online_params,
        'active_target_params': self.active_target_params,
        'active_optimizer_state': self.active_optimizer_state,
        'passive_online_params': self.passive_online_params,
        'passive_target_params': self.passive_target_params,
        'passive_optimizer_state': self.passive_optimizer_state,
    }
    return bundle_dictionary

  def unbundle(self, checkpoint_dir, iteration_number, bundle_dictionary):
    """Restores the agent from a checkpoint.

    Restores the agent's Python objects to those specified in bundle_dictionary,
    and restores the TensorFlow objects to those specified in the
    checkpoint_dir. If the checkpoint_dir does not exist, will not reset the
      agent's state.

    Args:
      checkpoint_dir: str, path to the checkpoint saved.
      iteration_number: int, checkpoint version, used when restoring the replay
        buffer.
      bundle_dictionary: dict, containing additional Python objects owned by
        the agent.

    Returns:
      bool, True if unbundling was successful.
    """
    try:
      # self._replay.load() will throw a NotFoundError if it does not find all
      # the necessary files.
      self._replay.load(checkpoint_dir, iteration_number)
    except tf.errors.NotFoundError:
      if not self.allow_partial_reload:
        # If we don't allow partial reloads, we will return False.
        return False
      logging.warning('Unable to reload replay buffer!')
    if bundle_dictionary is not None:
      self.state = bundle_dictionary['state']
      self.training_steps = bundle_dictionary['training_steps']
      self.active_online_params = bundle_dictionary['active_online_params']
      self.passive_online_params = bundle_dictionary['passive_online_params']
      self.active_target_params = bundle_dictionary['active_target_params']
      self.passive_target_params = bundle_dictionary['passive_target_params']
      # We recreate the optimizer with the new online weights.
      self.active_optimizer_state = bundle_dictionary['active_optimizer_state']
      self.passive_optimizer_state = (
          bundle_dictionary['passive_optimizer_state'])
    elif not self.allow_partial_reload:
      return False
    else:
      logging.warning("Unable to reload the agent's parameters!")
    return True



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Entry point for Atari 100k experiments.

"""

import functools
import json
import os

from absl import app
from absl import flags
from absl import logging
from dopamine.discrete_domains import run_experiment
from dopamine.discrete_domains import train as base_train
from dopamine.labs.atari_100k import atari_100k_rainbow_agent
from dopamine.labs.atari_100k import eval_run_experiment
import numpy as np
import tensorflow as tf


FLAGS = flags.FLAGS
AGENTS = ['DER', 'DrQ', 'OTRainbow', 'DrQ_eps']

# flags are defined when importing run_xm_preprocessing
flags.DEFINE_enum('agent', 'DER', AGENTS, 'Name of the agent.')
flags.DEFINE_integer('run_number', 1, 'Run number.')
flags.DEFINE_boolean('max_episode_eval', True,
                     'Whether to use `MaxEpisodeEvalRunner` or not.')


def create_agent(sess,  # pylint: disable=unused-argument
                 environment,
                 seed,
                 summary_writer=None):
  """Helper function for creating full rainbow-based Atari 100k agent."""
  return atari_100k_rainbow_agent.Atari100kRainbowAgent(
      num_actions=environment.action_space.n,
      seed=seed,
      summary_writer=summary_writer)


def set_random_seed(seed):
  """Set random seed for reproducibility."""
  logging.info('Setting random seed: %d', seed)
  os.environ['PYTHONHASHSEED'] = str(seed)
  tf.random.set_seed(seed)
  np.random.seed(seed)


def main(unused_argv):
  """Main method.

  Args:
    unused_argv: Arguments (unused).
  """
  logging.set_verbosity(logging.INFO)
  tf.compat.v1.disable_v2_behavior()
  base_dir = FLAGS.base_dir
  gin_files, gin_bindings = FLAGS.gin_files, FLAGS.gin_bindings
  run_experiment.load_gin_configs(gin_files, gin_bindings)
  # Set the Jax agent seed using the run number
  create_agent_fn = functools.partial(create_agent, seed=FLAGS.run_number)
  if FLAGS.max_episode_eval:
    runner_fn = eval_run_experiment.MaxEpisodeEvalRunner
    logging.info('Using MaxEpisodeEvalRunner for evaluation.')
    runner = runner_fn(base_dir, create_agent_fn)
  else:
    runner = run_experiment.Runner(base_dir, create_agent_fn)
  runner.run_experiment()


if __name__ == '__main__':
  flags.mark_flag_as_required('base_dir')
  app.run(main)



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Networks for continuous control agents."""

import functools
import operator
from typing import NamedTuple, Optional, Tuple

from flax import linen as nn
import gin
import jax
from jax import numpy as jnp
import tensorflow as tf
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions
tfb = tfp.bijectors


class SacActorOutput(NamedTuple):
  """The output of a SAC actor."""
  mean_action: jnp.ndarray
  sampled_action: jnp.ndarray
  log_probability: jnp.ndarray


class SacCriticOutput(NamedTuple):
  """The output of a SAC critic."""
  q_value1: jnp.ndarray
  q_value2: jnp.ndarray


class SacOutput(NamedTuple):
  """The output of a SACNetwork, including the actor and critic outputs."""
  actor: SacActorOutput
  critic: SacCriticOutput


class _Tanh(tfb.Tanh):

  def _inverse(self, y):
    # We perform clipping in the _inverse function, as is done in TF-Agents.
    y = jnp.where(
        jnp.less_equal(jnp.abs(y), 1.), tf.clip(y, -0.99999997, 0.99999997), y)
    return jnp.arctanh(y)


def _transform_distribution(dist, mean, magnitude):
  """Scales the input normal distribution to be within the action limits.

  Args:
    dist: a TensorFlow distribution.
    mean: desired action means.
    magnitude: desired action magnitudes.

  Returns:
    A transformed distribution, scaled to within the action limits.
  """
  bijectors = tfb.Chain([
      tfb.Shift(mean)(tfb.Scale(magnitude)),
      _Tanh(),
  ])
  dist = tfd.TransformedDistribution(dist, bijectors)
  return dist


def _shifted_uniform(minval=0., maxval=1.0, dtype=jnp.float32):

  def init(key, shape, dtype=dtype):
    return jax.random.uniform(
        key, shape=shape, minval=minval, maxval=maxval, dtype=dtype)

  return init


class SACCriticNetwork(nn.Module):
  """A simple critic network used in SAC."""
  num_layers: int = 2
  hidden_units: int = 256

  @nn.compact
  def __call__(self, state: jnp.ndarray, action: jnp.ndarray) -> jnp.ndarray:
    kernel_initializer = jax.nn.initializers.glorot_uniform()

    # Preprocess inputs
    a = action.reshape(-1)  # flatten
    x = state.astype(jnp.float32)
    x = x.reshape(-1)  # flatten
    x = jnp.concatenate((x, a))

    for _ in range(self.num_layers):
      x = nn.Dense(
          features=self.hidden_units, kernel_init=kernel_initializer)(
              x)
      x = nn.relu(x)

    return nn.Dense(features=1, kernel_init=kernel_initializer)(x)


@gin.configurable
class SACNetwork(nn.Module):
  """Non-convolutional value and policy networks for SAC."""
  action_shape: Tuple[int, ...]
  num_layers: int = 2
  hidden_units: int = 256
  action_limits: Optional[Tuple[Tuple[float, ...], Tuple[float, ...]]] = None

  def setup(self):
    action_dim = functools.reduce(operator.mul, self.action_shape, 1)

    # The setting of these initializers were borrowed from the TF-Agents SAC
    # implementation.
    kernel_initializer = jax.nn.initializers.glorot_uniform()

    self._critic1 = SACCriticNetwork(self.num_layers, self.hidden_units)
    self._critic2 = SACCriticNetwork(self.num_layers, self.hidden_units)

    self._actor_layers = [
        nn.Dense(features=self.hidden_units, kernel_init=kernel_initializer)
        for _ in range(self.num_layers)
    ]
    self._actor_final_layer = nn.Dense(
        features=action_dim * 2, kernel_init=kernel_initializer)

  def __call__(self,
               state: jnp.ndarray,
               key: jnp.ndarray,
               mean_action: bool = True) -> SacOutput:
    """Calls the SAC actor/critic networks.

    This has two important purposes:
      1. It is used to initialize all parameters of both networks.
      2. It is used to efficiently calculate the outputs of both the actor
        and critic networks on a single input.

    Args:
      state: An input state.
      key: A PRNGKey to use to sample an action from the actor's output
        distribution.
      mean_action: If True, it will use the actor's mean action to feed to the
        value network. Otherwise, it will use the sampled action.

    Returns:
      A named tuple containing the outputs from both networks.
    """
    actor_output = self.actor(state, key)

    if mean_action:
      critic_output = self.critic(state, actor_output.mean_action)
    else:
      critic_output = self.critic(state, actor_output.sampled_action)

    return SacOutput(actor_output, critic_output)

  def actor(self, state: jnp.ndarray, key: jnp.ndarray) -> SacActorOutput:
    """Calls the SAC actor network.

    This can be called using network_def.apply(..., method=network_def.actor).

    Args:
      state: An input state.
      key: A PRNGKey to use to sample an action from the actor's output
        distribution.

    Returns:
      A named tuple containing a sampled action, the mean action, and the
        likelihood of the sampled action.
    """
    # Preprocess inputs
    x = state.astype(jnp.float32)
    x = x.reshape(-1)  # flatten

    for layer in self._actor_layers:
      x = layer(x)
      x = nn.relu(x)

    # Note we are only producing a diagonal covariance matrix, not a full
    # covariance matrix as it is difficult to ensure that it would be PSD.
    loc_and_scale_diag = self._actor_final_layer(x)
    loc, scale_diag = jnp.split(loc_and_scale_diag, 2)
    # Exponentiate to only get positive terms.
    scale_diag = jnp.exp(scale_diag)
    dist = tfd.MultivariateNormalDiag(loc=loc, scale_diag=scale_diag)
    if self.action_limits is None:
      mode = dist.mode()
    else:
      lower_action_limit = jnp.asarray(self.action_limits[0], dtype=jnp.float32)
      upper_action_limit = jnp.asarray(self.action_limits[1], dtype=jnp.float32)
      mean = (lower_action_limit + upper_action_limit) / 2.0
      magnitude = (upper_action_limit - lower_action_limit) / 2.0
      mode = magnitude * jnp.tanh(dist.mode()) + mean
      dist = _transform_distribution(dist, mean, magnitude)
    sampled_action = dist.sample(seed=key)
    action_probability = dist.log_prob(sampled_action)

    mode = jnp.reshape(mode, self.action_shape)
    sampled_action = jnp.reshape(sampled_action, self.action_shape)

    return SacActorOutput(mode, sampled_action, action_probability)

  def critic(self, state: jnp.ndarray, action: jnp.ndarray) -> SacCriticOutput:
    """Calls the SAC critic network.

    SAC uses two Q networks to reduce overestimation bias.
    This can be called using network_def.apply(..., method=network_def.critic).

    Args:
      state: An input state.
      action: An action to compute the value of.

    Returns:
      A named tuple containing the Q values of each network.
    """
    return SacCriticOutput(
        self._critic1(state, action), self._critic2(state, action))



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Runner for evaluation based on a fixed number of episodes."""

import sys

from absl import logging
from dopamine.discrete_domains import run_experiment
import gin
import jax


@gin.configurable
class MaxEpisodeEvalRunner(run_experiment.Runner):
  """Runner for evaluating using a fixed number of episodes rather than steps."""

  def __init__(self,
               base_dir,
               create_agent_fn,
               num_eval_episodes=100,
               max_noops=30):
    """Specify the number of evaluation episodes."""
    super().__init__(base_dir, create_agent_fn)
    self._num_eval_episodes = num_eval_episodes
    logging.info('Num evaluation episodes: %d', num_eval_episodes)
    self._evaluation_steps = None
    self._max_noops = max_noops

  def _initialize_episode(self):
    """Initialization for a new episode with random number of no-ops.

    Returns:
     action: int, the initial action chosen by the agent.
    """
    initial_observation = self._environment.reset()
    if self._max_noops > 0:
      initial_observation = self._run_no_ops()
    return self._agent.begin_episode(initial_observation)

  def _run_no_ops(self):
    """Executes `num_noops` no-ops in the environment."""
    self._agent._rng, rng = jax.random.split(self._agent._rng)  # pylint: disable=protected-access
    num_noops = jax.random.randint(
        rng, shape=(), minval=0, maxval=self._max_noops)
    for _ in range(num_noops):
      # Assumes raw action 0 is always no-op
      self._environment.environment.ale.act(0)
    if self._environment.environment.ale.game_over():
      observation = self._environment.reset()
    else:
      observation = self._environment._pool_and_resize()  # pylint: disable=protected-access
    return observation

  def _run_one_phase_fix_episodes(self, max_episodes, statistics, run_mode_str):
    """Runs the agent/environment loop until a desired number of steps.

    We follow the Machado et al., 2017 convention of running full episodes,
    and terminating once we've run a minimum number of steps.

    Args:
      max_episodes: int, fixed number of episodes to generate in this phase.
      statistics: `IterationStatistics` object which records the experimental
        results.
      run_mode_str: str, describes the run mode for this agent.

    Returns:
      Tuple containing the number of steps taken in this phase (int), the sum of
        returns (float), and the number of episodes performed (int).
    """
    step_count = 0
    num_episodes = 0
    sum_returns = 0.

    while num_episodes < max_episodes:
      episode_length, episode_return = self._run_one_episode()
      statistics.append({
          '{}_episode_lengths'.format(run_mode_str): episode_length,
          '{}_episode_returns'.format(run_mode_str): episode_return
      })
      step_count += episode_length
      sum_returns += episode_return
      num_episodes += 1
      # We use sys.stdout.write instead of logging so as to flush frequently
      # without generating a line break.
      sys.stdout.write('Steps executed: {} '.format(step_count) +
                       'Episode length: {} '.format(episode_length) +
                       'Num episodes: {} '.format(num_episodes) +
                       'Return: {}\r'.format(episode_return))
      sys.stdout.flush()
    return step_count, sum_returns, num_episodes

  def _run_eval_phase(self, statistics):
    """Run evaluation phase.

    Args:
      statistics: `IterationStatistics` object which records the experimental
        results. Note - This object is modified by this method.

    Returns:
      num_episodes: int, The number of episodes run in this phase.
      average_reward: float, The average reward generated in this phase.
    """
    # Perform the evaluation phase -- no learning.
    self._agent.eval_mode = True
    _, sum_returns, num_episodes = self._run_one_phase_fix_episodes(
        self._num_eval_episodes, statistics, 'eval')
    average_return = sum_returns / num_episodes if num_episodes > 0 else 0.0
    logging.info('Average undiscounted return per evaluation episode: %.2f',
                 average_return)
    statistics.append({'eval_average_return': average_return})
    return num_episodes, average_return



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Various networks for Jax Dopamine agents."""

import time
from typing import Tuple, Union

from dopamine.discrete_domains import atari_lib
from flax import linen as nn
import gin
import jax
import jax.numpy as jnp
import numpy as onp


gin.constant('jax_networks.CARTPOLE_OBSERVATION_DTYPE', jnp.float64)
gin.constant('jax_networks.CARTPOLE_MIN_VALS',
             (-2.4, -5., -onp.pi/12., -onp.pi*2.))
gin.constant('jax_networks.CARTPOLE_MAX_VALS',
             (2.4, 5., onp.pi/12., onp.pi*2.))
gin.constant('jax_networks.ACROBOT_OBSERVATION_DTYPE', jnp.float64)
gin.constant('jax_networks.ACROBOT_MIN_VALS',
             (-1., -1., -1., -1., -5., -5.))
gin.constant('jax_networks.ACROBOT_MAX_VALS',
             (1., 1., 1., 1., 5., 5.))
gin.constant('jax_networks.LUNAR_OBSERVATION_DTYPE', jnp.float64)
gin.constant('jax_networks.MOUNTAINCAR_OBSERVATION_DTYPE', jnp.float64)
gin.constant('jax_networks.MOUNTAINCAR_MIN_VALS', (-1.2, -0.07))
gin.constant('jax_networks.MOUNTAINCAR_MAX_VALS', (0.6, 0.07))


def preprocess_atari_inputs(x):
  """Input normalization for Atari 2600 input frames."""
  return x.astype(jnp.float32) / 255.


identity_preprocess_fn = lambda x: x


### DQN Networks ###
@gin.configurable
class NatureDQNNetwork(nn.Module):
  """The convolutional network used to compute the agent's Q-values."""
  num_actions: int
  inputs_preprocessed: bool = False

  @nn.compact
  def __call__(self, x):
    initializer = nn.initializers.xavier_uniform()
    if not self.inputs_preprocessed:
      x = preprocess_atari_inputs(x)
    x = nn.Conv(features=32, kernel_size=(8, 8), strides=(4, 4),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(4, 4), strides=(2, 2),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(3, 3), strides=(1, 1),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = x.reshape((-1))  # flatten
    x = nn.Dense(features=512, kernel_init=initializer)(x)
    x = nn.relu(x)
    q_values = nn.Dense(features=self.num_actions,
                        kernel_init=initializer)(x)
    return atari_lib.DQNNetworkType(q_values)


@gin.configurable
class ClassicControlDQNNetwork(nn.Module):
  """Jax DQN network for classic control environments."""
  num_actions: int
  num_layers: int = 2
  hidden_units: int = 512
  min_vals: Union[None, Tuple[float, ...]] = None
  max_vals: Union[None, Tuple[float, ...]] = None
  inputs_preprocessed: bool = False

  def setup(self):
    if self.min_vals is not None:
      assert self.max_vals is not None
      self._min_vals = jnp.array(self.min_vals)
      self._max_vals = jnp.array(self.max_vals)
    initializer = nn.initializers.xavier_uniform()
    self.layers = [
        nn.Dense(features=self.hidden_units, kernel_init=initializer)
        for _ in range(self.num_layers)]
    self.final_layer = nn.Dense(features=self.num_actions,
                                kernel_init=initializer)

  def __call__(self, x):
    if not self.inputs_preprocessed:
      x = x.astype(jnp.float32)
      x = x.reshape((-1))  # flatten
      if self.min_vals is not None:
        x -= self._min_vals
        x /= self._max_vals - self._min_vals
        x = 2.0 * x - 1.0  # Rescale in range [-1, 1].
    for layer in self.layers:
      x = layer(x)
      x = nn.relu(x)
    q_values = self.final_layer(x)
    return atari_lib.DQNNetworkType(q_values)


### Rainbow Networks ###
@gin.configurable
class RainbowNetwork(nn.Module):
  """Convolutional network used to compute the agent's return distributions."""
  num_actions: int
  num_atoms: int
  inputs_preprocessed: bool = False

  @nn.compact
  def __call__(self, x, support):
    initializer = nn.initializers.variance_scaling(
        scale=1.0 / jnp.sqrt(3.0),
        mode='fan_in',
        distribution='uniform')
    if not self.inputs_preprocessed:
      x = preprocess_atari_inputs(x)
    x = nn.Conv(features=32, kernel_size=(8, 8), strides=(4, 4),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(4, 4), strides=(2, 2),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(3, 3), strides=(1, 1),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = x.reshape((-1))  # flatten
    x = nn.Dense(features=512, kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Dense(features=self.num_actions * self.num_atoms,
                 kernel_init=initializer)(x)
    logits = x.reshape((self.num_actions, self.num_atoms))
    probabilities = nn.softmax(logits)
    q_values = jnp.sum(support * probabilities, axis=1)
    return atari_lib.RainbowNetworkType(q_values, logits, probabilities)


@gin.configurable
class ClassicControlRainbowNetwork(nn.Module):
  """Jax Rainbow network for classic control environments."""
  num_actions: int
  num_atoms: int
  num_layers: int = 2
  hidden_units: int = 512
  min_vals: Union[None, Tuple[float, ...]] = None
  max_vals: Union[None, Tuple[float, ...]] = None
  inputs_preprocessed: bool = False

  def setup(self):
    if self.min_vals is not None:
      self._min_vals = jnp.array(self.min_vals)
      self._max_vals = jnp.array(self.max_vals)
    initializer = nn.initializers.xavier_uniform()
    self.layers = [
        nn.Dense(features=self.hidden_units, kernel_init=initializer)
        for _ in range(self.num_layers)]
    self.final_layer = nn.Dense(features=self.num_actions * self.num_atoms,
                                kernel_init=initializer)

  def __call__(self, x, support):
    if not self.inputs_preprocessed:
      x = x.astype(jnp.float32)
      x = x.reshape((-1))  # flatten
      if self.min_vals is not None:
        x -= self._min_vals
        x /= self._max_vals - self._min_vals
        x = 2.0 * x - 1.0  # Rescale in range [-1, 1].
    for layer in self.layers:
      x = layer(x)
      x = nn.relu(x)
    x = self.final_layer(x)
    logits = x.reshape((self.num_actions, self.num_atoms))
    probabilities = nn.softmax(logits)
    q_values = jnp.sum(support * probabilities, axis=1)
    return atari_lib.RainbowNetworkType(q_values, logits, probabilities)


### Implicit Quantile Networks ###
class ImplicitQuantileNetwork(nn.Module):
  """The Implicit Quantile Network (Dabney et al., 2018).."""
  num_actions: int
  quantile_embedding_dim: int
  inputs_preprocessed: bool = False

  @nn.compact
  def __call__(self, x, num_quantiles, rng):
    initializer = nn.initializers.variance_scaling(
        scale=1.0 / jnp.sqrt(3.0),
        mode='fan_in',
        distribution='uniform')
    if not self.inputs_preprocessed:
      x = preprocess_atari_inputs(x)
    x = nn.Conv(features=32, kernel_size=(8, 8), strides=(4, 4),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(4, 4), strides=(2, 2),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(3, 3), strides=(1, 1),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = x.reshape((-1))  # flatten
    state_vector_length = x.shape[-1]
    state_net_tiled = jnp.tile(x, [num_quantiles, 1])
    quantiles_shape = [num_quantiles, 1]
    quantiles = jax.random.uniform(rng, shape=quantiles_shape)
    quantile_net = jnp.tile(quantiles, [1, self.quantile_embedding_dim])
    quantile_net = (
        jnp.arange(1, self.quantile_embedding_dim + 1, 1).astype(jnp.float32)
        * onp.pi
        * quantile_net)
    quantile_net = jnp.cos(quantile_net)
    quantile_net = nn.Dense(features=state_vector_length,
                            kernel_init=initializer)(quantile_net)
    quantile_net = nn.relu(quantile_net)
    x = state_net_tiled * quantile_net
    x = nn.Dense(features=512, kernel_init=initializer)(x)
    x = nn.relu(x)
    quantile_values = nn.Dense(features=self.num_actions,
                               kernel_init=initializer)(x)
    return atari_lib.ImplicitQuantileNetworkType(quantile_values, quantiles)


### Quantile Networks ###
@gin.configurable
class QuantileNetwork(nn.Module):
  """Convolutional network used to compute the agent's return quantiles."""
  num_actions: int
  num_atoms: int
  inputs_preprocessed: bool = False

  @nn.compact
  def __call__(self, x):
    initializer = nn.initializers.variance_scaling(
        scale=1.0 / jnp.sqrt(3.0),
        mode='fan_in',
        distribution='uniform')
    if not self.inputs_preprocessed:
      x = preprocess_atari_inputs(x)
    x = nn.Conv(features=32, kernel_size=(8, 8), strides=(4, 4),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(4, 4), strides=(2, 2),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Conv(features=64, kernel_size=(3, 3), strides=(1, 1),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = x.reshape((-1))  # flatten
    x = nn.Dense(features=512, kernel_init=initializer)(x)
    x = nn.relu(x)
    x = nn.Dense(features=self.num_actions * self.num_atoms,
                 kernel_init=initializer)(x)
    logits = x.reshape((self.num_actions, self.num_atoms))
    probabilities = nn.softmax(logits)
    q_values = jnp.mean(logits, axis=1)
    return atari_lib.RainbowNetworkType(q_values, logits, probabilities)


### Noisy Nets for FullRainbowNetwork ###
@gin.configurable
class NoisyNetwork(nn.Module):
  """Noisy Network from Fortunato et al. (2018).

  Attributes:
    rng_key: jax.interpreters.xla.DeviceArray, key for JAX RNG.
    eval_mode: bool, whether to turn off noise during evaluation.
  """
  rng_key: jax.interpreters.xla.DeviceArray
  eval_mode: bool = False

  @staticmethod
  def sample_noise(key, shape):
    return jax.random.normal(key, shape)

  @staticmethod
  def f(x):
    # See (10) and (11) in Fortunato et al. (2018).
    return jnp.multiply(jnp.sign(x), jnp.power(jnp.abs(x), 0.5))

  @nn.compact
  def __call__(self, x, features, bias=True, kernel_init=None):

    def mu_init(key, shape):
      # Initialization of mean noise parameters (Section 3.2)
      low = -1 / jnp.power(x.shape[0], 0.5)
      high = 1 / jnp.power(x.shape[0], 0.5)
      return jax.random.uniform(key, minval=low, maxval=high, shape=shape)

    def sigma_init(key, shape, dtype=jnp.float32):  # pylint: disable=unused-argument
      # Initialization of sigma noise parameters (Section 3.2)
      return jnp.ones(shape, dtype) * (0.1 / onp.sqrt(x.shape[0]))

    if self.eval_mode:
      # Turn off noise during evaluation
      w_epsilon = onp.zeros(shape=(x.shape[0], features), dtype=onp.float32)
      b_epsilon = onp.zeros(shape=(features,), dtype=onp.float32)
    else:
      # Factored gaussian noise in (10) and (11) in Fortunato et al. (2018).
      p = NoisyNetwork.sample_noise(self.rng_key, [x.shape[0], 1])
      q = NoisyNetwork.sample_noise(self.rng_key, [1, features])
      f_p = NoisyNetwork.f(p)
      f_q = NoisyNetwork.f(q)
      w_epsilon = f_p * f_q
      b_epsilon = jnp.squeeze(f_q)

    # See (8) and (9) in Fortunato et al. (2018) for output computation.
    w_mu = self.param('kernel_mu', mu_init, (x.shape[0], features))
    w_sigma = self.param('kernel_sigma', sigma_init, (x.shape[0], features))
    w = w_mu + jnp.multiply(w_sigma, w_epsilon)
    ret = jnp.matmul(x, w)

    b_mu = self.param('bias_mu', mu_init, (features,))
    b_sigma = self.param('bias_sigma', sigma_init, (features,))
    b = b_mu + jnp.multiply(b_sigma, b_epsilon)
    return jnp.where(bias, ret + b, ret)


### FullRainbowNetwork ###
def feature_layer(key, noisy, eval_mode=False):
  """Network feature layer depending on whether noisy_nets are used on or not."""
  def noisy_net(x, features):
    return NoisyNetwork(rng_key=key, eval_mode=eval_mode)(x, features)
  def dense_net(x, features):
    return nn.Dense(features, kernel_init=nn.initializers.xavier_uniform())(x)
  return noisy_net if noisy else dense_net


@gin.configurable
class FullRainbowNetwork(nn.Module):
  """Jax Rainbow network for Full Rainbow.

  Attributes:
    num_actions: int, number of actions the agent can take at any state.
    num_atoms: int, the number of buckets of the value function distribution.
    noisy: bool, Whether to use noisy networks.
    dueling: bool, Whether to use dueling network architecture.
    distributional: bool, whether to use distributional RL.
  """
  num_actions: int
  num_atoms: int
  noisy: bool = True
  dueling: bool = True
  distributional: bool = True
  inputs_preprocessed: bool = False

  @nn.compact
  def __call__(self, x, support, eval_mode=False, key=None):
    # Generate a random number generation key if not provided
    if key is None:
      key = jax.random.PRNGKey(int(time.time() * 1e6))

    if not self.inputs_preprocessed:
      x = preprocess_atari_inputs(x)

    hidden_sizes = [32, 64, 64]
    kernel_sizes = [8, 4, 3]
    stride_sizes = [4, 2, 1]
    for hidden_size, kernel_size, stride_size in zip(hidden_sizes, kernel_sizes,
                                                     stride_sizes):
      x = nn.Conv(
          features=hidden_size,
          kernel_size=(kernel_size, kernel_size),
          strides=(stride_size, stride_size),
          kernel_init=nn.initializers.xavier_uniform())(x)
      x = nn.relu(x)
    x = x.reshape((-1))  # flatten

    net = feature_layer(key, self.noisy, eval_mode=eval_mode)
    x = net(x, features=512)  # Single hidden layer of size 512
    x = nn.relu(x)

    if self.dueling:
      adv = net(x, features=self.num_actions * self.num_atoms)
      value = net(x, features=self.num_atoms)
      adv = adv.reshape((self.num_actions, self.num_atoms))
      value = value.reshape((1, self.num_atoms))
      logits = value + (adv - (jnp.mean(adv, axis=0, keepdims=True)))
    else:
      x = net(x, features=self.num_actions * self.num_atoms)
      logits = x.reshape((self.num_actions, self.num_atoms))

    if self.distributional:
      probabilities = nn.softmax(logits)
      q_values = jnp.sum(support * probabilities, axis=1)
      return atari_lib.RainbowNetworkType(q_values, logits, probabilities)
    q_values = jnp.sum(logits, axis=1)  # Sum over all the num_atoms
    return atari_lib.DQNNetworkType(q_values)



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module defining classes and helper methods for general agents."""

from typing import Optional

from absl import logging
from dopamine.discrete_domains import gym_lib
from dopamine.discrete_domains import iteration_statistics
from dopamine.discrete_domains import run_experiment as base_run_experiment
from dopamine.jax.agents.dqn import dqn_agent
from dopamine.jax.agents.sac import sac_agent
# pylint: disable=unused-import
from dopamine.labs.sac_from_pixels import deepmind_control_lib
# pylint: enable=unused-import
from flax.metrics import tensorboard
import gin
from gym import spaces



load_gin_configs = base_run_experiment.load_gin_configs


@gin.configurable
def create_continuous_agent(
    environment: gym_lib.GymPreprocessing,
    agent_name: str,
    summary_writer: Optional[tensorboard.SummaryWriter] = None
) -> dqn_agent.JaxDQNAgent:
  """Creates an agent.

  Args:
    environment:  A gym environment.
    agent_name: str, name of the agent to create.
    summary_writer: A Tensorflow summary writer to pass to the agent
      for in-agent training statistics in Tensorboard.

  Returns:
    An RL agent.

  Raises:
    ValueError: If `agent_name` is not in supported list.
  """
  assert agent_name is not None
  if agent_name == 'sac':
    assert isinstance(environment.action_space, spaces.Box)
    assert isinstance(environment.observation_space, spaces.Box)
    return sac_agent.SACAgent(
        action_shape=environment.action_space.shape,
        action_limits=(environment.action_space.low,
                       environment.action_space.high),
        observation_shape=environment.observation_space.shape,
        action_dtype=environment.action_space.dtype,
        observation_dtype=environment.observation_space.dtype,
        summary_writer=summary_writer)
  else:
    raise ValueError(f'Unknown agent: {agent_name}')


@gin.configurable
def create_continuous_runner(base_dir, schedule='continuous_train_and_eval'):
  """Creates an experiment Runner.

  Args:
    base_dir: str, base directory for hosting all subdirectories.
    schedule: string, which type of Runner to use.

  Returns:
    runner: A `Runner` like object.

  Raises:
    ValueError: When an unknown schedule is encountered.
  """
  assert base_dir is not None
  # Continuously runs training and evaluation until max num_iterations is hit.
  if schedule == 'continuous_train_and_eval':
    return ContinuousRunner(base_dir, create_continuous_agent)
  # Continuously runs training until max num_iterations is hit.
  elif schedule == 'continuous_train':
    return ContinuousTrainRunner(base_dir, create_continuous_agent)
  else:
    raise ValueError('Unknown schedule: {}'.format(schedule))


@gin.configurable
class ContinuousRunner(base_run_experiment.Runner):
  """Object that handles running Dopamine experiments.

  This is mostly the same as discrete_domains.Runner, but is written solely for
  JAX/Flax agents.
  """

  def __init__(self,
               base_dir,
               create_agent_fn,
               create_environment_fn=gym_lib.create_gym_environment,
               checkpoint_file_prefix='ckpt',
               logging_file_prefix='log',
               log_every_n=1,
               num_iterations=200,
               training_steps=250000,
               evaluation_steps=125000,
               max_steps_per_episode=1000,
               clip_rewards=False):
    """Initialize the Runner object in charge of running a full experiment.

    Args:
      base_dir: str, the base directory to host all required sub-directories.
      create_agent_fn: A function that takes as argument an environment, and
        returns an agent.
      create_environment_fn: A function which receives a problem name and
        creates a Gym environment for that problem (e.g. an Atari 2600 game).
      checkpoint_file_prefix: str, the prefix to use for checkpoint files.
      logging_file_prefix: str, prefix to use for the log files.
      log_every_n: int, the frequency for writing logs.
      num_iterations: int, the iteration number threshold (must be greater than
        start_iteration).
      training_steps: int, the number of training steps to perform.
      evaluation_steps: int, the number of evaluation steps to perform.
      max_steps_per_episode: int, maximum number of steps after which an episode
        terminates.
      clip_rewards: bool, whether to clip rewards in [-1, 1].

    This constructor will take the following actions:
    - Initialize an environment.
    - Initialize a logger.
    - Initialize an agent.
    - Reload from the latest checkpoint, if available, and initialize the
      Checkpointer object.
    """
    assert base_dir is not None
    self._logging_file_prefix = logging_file_prefix
    self._log_every_n = log_every_n
    self._num_iterations = num_iterations
    self._training_steps = training_steps
    self._evaluation_steps = evaluation_steps
    self._max_steps_per_episode = max_steps_per_episode
    self._base_dir = base_dir
    self._clip_rewards = clip_rewards
    self._create_directories()
    self._summary_writer = tensorboard.SummaryWriter(base_dir)
    self._environment = create_environment_fn()
    self._agent = create_agent_fn(self._environment,
                                  summary_writer=self._summary_writer)
    self._initialize_checkpointer_and_maybe_resume(checkpoint_file_prefix)

  def _save_tensorboard_summaries(self, iteration,
                                  num_episodes_train,
                                  average_reward_train,
                                  num_episodes_eval,
                                  average_reward_eval,
                                  average_steps_per_second):
    """Save statistics as tensorboard summaries.

    Args:
      iteration: int, The current iteration number.
      num_episodes_train: int, number of training episodes run.
      average_reward_train: float, The average training reward.
      num_episodes_eval: int, number of evaluation episodes run.
      average_reward_eval: float, The average evaluation reward.
      average_steps_per_second: float, The average number of steps per second.
    """
    metrics = [('Train/NumEpisodes', num_episodes_train),
               ('Train/AverageReturns', average_reward_train),
               ('Train/AverageStepsPerSecond', average_steps_per_second),
               ('Eval/NumEpisodes', num_episodes_eval),
               ('Eval/AverageReturns', average_reward_eval)]
    for name, value in metrics:
      self._summary_writer.scalar(name, value, iteration)
    self._summary_writer.flush()


@gin.configurable
class ContinuousTrainRunner(ContinuousRunner):
  """Object that handles running experiments.

  This is mostly the same as discrete_domains.TrainRunner, but is written solely
  for JAX/Flax agents.
  """

  def __init__(self, base_dir, create_agent_fn,
               create_environment_fn=gym_lib.create_gym_environment):
    """Initialize the TrainRunner object in charge of running a full experiment.

    Args:
      base_dir: str, the base directory to host all required sub-directories.
      create_agent_fn: A function that takes as args a Tensorflow session and an
        environment, and returns an agent.
      create_environment_fn: A function which receives a problem name and
        creates a Gym environment for that problem (e.g. an Atari 2600 game).
    """
    logging.info('Creating ContinuousTrainRunner ...')
    super().__init__(base_dir, create_agent_fn, create_environment_fn)
    self._agent.eval_mode = False

  def _run_one_iteration(self, iteration):
    """Runs one iteration of agent/environment interaction.

    An iteration involves running several episodes until a certain number of
    steps are obtained. This method differs from the `_run_one_iteration` method
    in the base `Runner` class in that it only runs the train phase.

    Args:
      iteration: int, current iteration number, used as a global_step for saving
        Tensorboard summaries.

    Returns:
      A dict containing summary statistics for this iteration.
    """
    statistics = iteration_statistics.IterationStatistics()
    num_episodes_train, average_reward_train, average_steps_per_second = (
        self._run_train_phase(statistics))

    self._save_tensorboard_summaries(iteration, num_episodes_train,
                                     average_reward_train,
                                     average_steps_per_second)
    return statistics.data_lists

  def _save_tensorboard_summaries(self, iteration, num_episodes,
                                  average_reward, average_steps_per_second):
    """Save statistics as tensorboard summaries."""
    metrics = [('Train/NumEpisodes', num_episodes),
               ('Train/AverageReturns', average_reward),
               ('Train/AverageStepsPerSecond', average_steps_per_second)]
    for name, value in metrics:
      self._summary_writer.scalar(name, value, iteration)
    self._summary_writer.flush()



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""The entry point for running a Dopamine agent on continuous control envs.

"""

from absl import app
from absl import flags
from absl import logging

from dopamine.continuous_domains import run_experiment

flags.DEFINE_string('base_dir', None,
                    'Base directory to host all required sub-directories.')
flags.DEFINE_multi_string(
    'gin_files', [], 'List of paths to gin configuration files (e.g.'
    '"dopamine/jax/agents/sac/configs/sac.gin").')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')

FLAGS = flags.FLAGS


def main(unused_argv):
  """Main method.

  Args:
    unused_argv: Arguments (unused).
  """
  logging.set_verbosity(logging.INFO)
  base_dir = FLAGS.base_dir
  gin_files = FLAGS.gin_files
  gin_bindings = FLAGS.gin_bindings

  run_experiment.load_gin_configs(gin_files, gin_bindings)
  runner = run_experiment.create_continuous_runner(base_dir)
  runner.run_experiment()


if __name__ == '__main__':
  flags.mark_flag_as_required('base_dir')
  app.run(main)



# coding=utf-8
# Copyright 2021 The Tandem DQN authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run training with Tandem-RL: one active and one passive agent."""

import sys

from absl import logging
from dopamine.discrete_domains import atari_lib
from dopamine.discrete_domains import gym_lib
from dopamine.discrete_domains import iteration_statistics
from dopamine.discrete_domains import run_experiment
from dopamine.labs.environments.minatar import minatar_env
from dopamine.labs.tandem_dqn import tandem_dqn_agent
import gin
import numpy as np
import tensorflow as tf


@gin.configurable
def create_tandem_agents_and_checkpoints(sess, environment, agent_name='dqn',
                                         summary_writer=None, debug_mode=False):
  """Creates a tandem agent.

  Args:
    sess: TF session, unused since we are in JAX.
    environment: A gym environment (e.g. Atari 2600).
    agent_name: str, name of the agent to create.
    summary_writer: A Tensorflow summary writer to pass to the agent
      for in-agent training statistics in Tensorboard.
    debug_mode: bool, unused.

  Returns:
    An active and passive agent.
  """
  assert agent_name is not None
  del sess
  del debug_mode
  if agent_name == 'dqn':
    return tandem_dqn_agent.TandemDQNAgent(
        num_actions=environment.action_space.n, summary_writer=summary_writer)
  else:
    raise ValueError('Unknown agent: {}'.format(agent_name))


@gin.configurable
class TandemRunner(run_experiment.Runner):
  """Runner class to run Tandem RL experiments."""

  def __init__(self,
               base_dir,
               create_agent_fn,
               suite='atari'):
    if suite == 'atari':
      create_environment_fn = atari_lib.create_atari_environment
    elif suite == 'classic':
      create_environment_fn = gym_lib.create_gym_environment
    elif suite == 'minatar':
      create_environment_fn = minatar_env.create_minatar_env
    else:
      raise ValueError(f'Unknown suite: {suite}')
    super().__init__(base_dir, create_agent_fn,
                     create_environment_fn=create_environment_fn)

  def _initialize_episode(self, agent_type='active'):
    """Initialization for a new episode.

    Args:
      agent_type: str, the type of agent to run.

    Returns:
      action: int, the initial action chosen by the agent.
    """
    initial_observation = self._environment.reset()
    return self._agent.begin_episode(agent_type, initial_observation)

  def _run_one_episode(self, agent_type='active'):
    """Executes a full trajectory of the agent interacting with the environment.

    Args:
      agent_type: str, the type of agent to run.

    Returns:
      The number of steps taken and the total reward.
    """
    step_number = 0
    total_reward = 0.

    action = self._initialize_episode(agent_type)
    is_terminal = False

    # Keep interacting until we reach a terminal state.
    while True:
      observation, reward, is_terminal = self._run_one_step(action)

      total_reward += reward
      step_number += 1

      # Perform reward clipping.
      reward = np.clip(reward, -1, 1)

      if (self._environment.game_over or
          step_number == self._max_steps_per_episode):
        # Stop the run loop once we reach the true end of episode.
        break
      elif is_terminal:
        # If we lose a life but the episode is not over, signal an artificial
        # end of episode to the agent.
        if agent_type == 'active':
          self._agent.end_episode(reward)
        action = self._agent.begin_episode(agent_type, observation)
      else:
        action = self._agent.step(agent_type, reward, observation)

    self._end_episode(reward)

    return step_number, total_reward

  def _run_one_phase(self, min_steps, statistics, run_mode_str,
                     agent_type='active'):
    """Runs the agent/environment loop until a desired number of steps.

    We follow the Machado et al., 2017 convention of running full episodes,
    and terminating once we've run a minimum number of steps.

    Args:
      min_steps: int, minimum number of steps to generate in this phase.
      statistics: `IterationStatistics` object which records the experimental
        results.
      run_mode_str: str, describes the run mode for this agent.
      agent_type: str, the type of agent to run.

    Returns:
      Tuple containing the number of steps taken in this phase (int), the sum of
        returns (float), and the number of episodes performed (int).
    """
    step_count = 0
    num_episodes = 0
    sum_returns = 0.

    while step_count < min_steps:
      episode_length, episode_return = self._run_one_episode(agent_type)
      statistics.append({
          '{}_{}_episode_lengths'.format(run_mode_str,
                                         agent_type): episode_length,
          '{}_{}_episode_returns'.format(run_mode_str,
                                         agent_type): episode_return
      })
      step_count += episode_length
      sum_returns += episode_return
      num_episodes += 1
      # We use sys.stdout.write instead of logging so as to flush frequently
      # without generating a line break.
      sys.stdout.write('Steps executed: {} '.format(step_count) +
                       'Episode length: {} '.format(episode_length) +
                       'Return: {}\r'.format(episode_return))
      sys.stdout.flush()
    return step_count, sum_returns, num_episodes

  def _run_eval_phase(self, statistics, agent_type='active'):
    """Run evaluation phase on both passive and active learners.

    Args:
      statistics: `IterationStatistics` object which records the experimental
        results. Note - This object is modified by this method.
      agent_type: str, the type of agent to run.

    Returns:
      num_episodes: int, The number of episodes run in this phase.
      average_reward: float, The average reward generated in this phase.
    """
    # Perform the evaluation phase -- no learning.
    self._agent.eval_mode = True
    _, sum_returns, num_episodes = self._run_one_phase(
        self._evaluation_steps, statistics, 'eval', agent_type)
    average_return = sum_returns / num_episodes if num_episodes > 0 else 0.0
    logging.info('Average undiscounted return per eval episode (%s): %.2f',
                 agent_type, average_return)
    statistics.append({f'{agent_type}_eval_average_return': average_return})
    return num_episodes, average_return

  def _run_one_iteration(self, iteration):
    """Runs one iteration of agent/environment interaction.

    An iteration involves running several episodes until a certain number of
    steps are obtained. The interleaving of train/eval phases implemented here
    are to match the implementation of (Mnih et al., 2015).

    Args:
      iteration: int, current iteration number, used as a global_step for saving
        Tensorboard summaries.

    Returns:
      A dict containing summary statistics for this iteration.
    """
    statistics = iteration_statistics.IterationStatistics()
    logging.info('Starting iteration %d', iteration)
    num_episodes_train, average_reward_train, average_steps_per_second = (
        self._run_train_phase(statistics))
    active_num_episodes_eval, active_average_reward_eval = self._run_eval_phase(
        statistics, 'active')
    passive_num_episodes_eval, passive_average_reward_eval = (
        self._run_eval_phase(statistics, 'passive'))

    self._save_tensorboard_summaries(iteration, num_episodes_train,
                                     average_reward_train,
                                     active_num_episodes_eval,
                                     active_average_reward_eval,
                                     passive_num_episodes_eval,
                                     passive_average_reward_eval,
                                     average_steps_per_second)
    return statistics.data_lists

  def _save_tensorboard_summaries(self, iteration,
                                  num_episodes_train,
                                  average_reward_train,
                                  active_num_episodes_eval,
                                  active_average_reward_eval,
                                  passive_num_episodes_eval,
                                  passive_average_reward_eval,
                                  average_steps_per_second):
    """Save statistics as tensorboard summaries.

    Args:
      iteration: int, The current iteration number.
      num_episodes_train: int, number of training episodes run.
      average_reward_train: float, The average training reward.
      active_num_episodes_eval: int, number of active evaluation episodes run.
      active_average_reward_eval: float, The average active evaluation reward.
      passive_num_episodes_eval: int, number of passive evaluation episodes run.
      passive_average_reward_eval: float, The average passive evaluation reward.
      average_steps_per_second: float, The average number of steps per second.
    """
    summary = tf.compat.v1.Summary(value=[
        tf.compat.v1.Summary.Value(
            tag='Train/NumEpisodes', simple_value=num_episodes_train),
        tf.compat.v1.Summary.Value(
            tag='Train/AverageReturns', simple_value=average_reward_train),
        tf.compat.v1.Summary.Value(
            tag='Train/AverageStepsPerSecond',
            simple_value=average_steps_per_second),
        tf.compat.v1.Summary.Value(
            tag='Eval/ActiveNumEpisodes',
            simple_value=active_num_episodes_eval),
        tf.compat.v1.Summary.Value(
            tag='Eval/ActiveAverageReturns',
            simple_value=active_average_reward_eval),
        tf.compat.v1.Summary.Value(
            tag='Eval/PassiveNumEpisodes',
            simple_value=passive_num_episodes_eval),
        tf.compat.v1.Summary.Value(
            tag='Eval/PassiveAverageReturns',
            simple_value=passive_average_reward_eval)
    ])
    self._summary_writer.add_summary(summary, iteration)
    self._summary_writer.flush()



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An extension of Rainbow to perform quantile regression.

This loss is computed as in "Distributional Reinforcement Learning with Quantile
Regression" - Dabney et. al, 2017"
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

from dopamine.jax import networks
from dopamine.jax.agents.dqn import dqn_agent
from dopamine.replay_memory import prioritized_replay_buffer
import gin
import jax
import jax.numpy as jnp
import optax
import tensorflow as tf


@functools.partial(jax.vmap, in_axes=(None, 0, 0, 0, None))
def target_distribution(target_network, next_states, rewards, terminals,
                        cumulative_gamma):
  """Builds the Quantile target distribution as per Dabney et al. (2017).

  Args:
    target_network: Jax Module used for the target network.
    next_states: numpy array of batched next states.
    rewards: numpy array of batched rewards.
    terminals: numpy array of batched terminals.
    cumulative_gamma: float, cumulative gamma to use.

  Returns:
    The target distribution from the replay.
  """
  is_terminal_multiplier = 1. - terminals.astype(jnp.float32)
  # Incorporate terminal state to discount factor.
  gamma_with_terminal = cumulative_gamma * is_terminal_multiplier
  next_state_target_outputs = target_network(next_states)
  q_values = jnp.squeeze(next_state_target_outputs.q_values)
  next_qt_argmax = jnp.argmax(q_values)
  logits = jnp.squeeze(next_state_target_outputs.logits)
  next_logits = logits[next_qt_argmax]
  return jax.lax.stop_gradient(rewards + gamma_with_terminal * next_logits)


@functools.partial(jax.jit, static_argnums=(0, 3, 10, 11, 12))
def train(network_def, online_params, target_params, optimizer, optimizer_state,
          states, actions, next_states, rewards, terminals, kappa, num_atoms,
          cumulative_gamma):
  """Run a training step."""
  def loss_fn(params, target):
    def q_online(state):
      return network_def.apply(params, state)

    logits = jax.vmap(q_online)(states).logits
    logits = jnp.squeeze(logits)
    # Fetch the logits for its selected action. We use vmap to perform this
    # indexing across the batch.
    chosen_action_logits = jax.vmap(lambda x, y: x[y])(logits, actions)
    bellman_errors = (target[:, None, :] -
                      chosen_action_logits[:, :, None])  # Input `u' of Eq. 9.
    # Eq. 9 of paper.
    huber_loss = (
        (jnp.abs(bellman_errors) <= kappa).astype(jnp.float32) *
        0.5 * bellman_errors ** 2 +
        (jnp.abs(bellman_errors) > kappa).astype(jnp.float32) *
        kappa * (jnp.abs(bellman_errors) - 0.5 * kappa))

    tau_hat = ((jnp.arange(num_atoms, dtype=jnp.float32) + 0.5) /
               num_atoms)  # Quantile midpoints.  See Lemma 2 of paper.
    # Eq. 10 of paper.
    tau_bellman_diff = jnp.abs(
        tau_hat[None, :, None] - (bellman_errors < 0).astype(jnp.float32))
    quantile_huber_loss = tau_bellman_diff * huber_loss
    # Sum over tau dimension, average over target value dimension.
    loss = jnp.sum(jnp.mean(quantile_huber_loss, 2), 1)
    return jnp.mean(loss), loss

  def q_target(state):
    return network_def.apply(target_params, state)

  grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
  target = target_distribution(q_target,
                               next_states,
                               rewards,
                               terminals,
                               cumulative_gamma)
  (mean_loss, loss), grad = grad_fn(online_params, target)
  updates, optimizer_state = optimizer.update(grad, optimizer_state,
                                              params=online_params)
  online_params = optax.apply_updates(online_params, updates)
  return optimizer_state, online_params, loss, mean_loss


@gin.configurable
class JaxQuantileAgent(dqn_agent.JaxDQNAgent):
  """An implementation of Quantile regression DQN agent."""

  def __init__(self,
               num_actions,
               observation_shape=dqn_agent.NATURE_DQN_OBSERVATION_SHAPE,
               observation_dtype=dqn_agent.NATURE_DQN_DTYPE,
               stack_size=dqn_agent.NATURE_DQN_STACK_SIZE,
               network=networks.QuantileNetwork,
               kappa=1.0,
               num_atoms=200,
               gamma=0.99,
               update_horizon=1,
               min_replay_history=50000,
               update_period=4,
               target_update_period=10000,
               epsilon_fn=dqn_agent.linearly_decaying_epsilon,
               epsilon_train=0.1,
               epsilon_eval=0.05,
               epsilon_decay_period=1000000,
               replay_scheme='prioritized',
               optimizer='adam',
               summary_writer=None,
               summary_writing_frequency=500,
               seed=None,
               allow_partial_reload=False):
    """Initializes the agent and constructs the Graph.

    Args:
      num_actions: Int, number of actions the agent can take at any state.
      observation_shape: tuple of ints or an int. If single int, the observation
        is assumed to be a 2D square.
      observation_dtype: DType, specifies the type of the observations. Note
        that if your inputs are continuous, you should set this to jnp.float32.
      stack_size: int, number of frames to use in state stack.
      network: flax.linen Module, expects 3 parameters: num_actions, num_atoms,
        network_type.
      kappa: Float, Huber loss cutoff.
      num_atoms: Int, the number of buckets for the value function distribution.
      gamma: Float, exponential decay factor as commonly used in the RL
        literature.
      update_horizon: Int, horizon at which updates are performed, the 'n' in
        n-step update.
      min_replay_history: Int, number of stored transitions for training to
        start.
      update_period: Int, period between DQN updates.
      target_update_period: Int, ppdate period for the target network.
      epsilon_fn: Function expecting 4 parameters: (decay_period, step,
        warmup_steps, epsilon), and which returns the epsilon value used for
        exploration during training.
      epsilon_train: Float, final epsilon for training.
      epsilon_eval: Float, epsilon during evaluation.
      epsilon_decay_period: Int, number of steps for epsilon to decay.
      replay_scheme: String, replay memory scheme to be used. Choices are:
        uniform - Standard (DQN) replay buffer (Mnih et al., 2015)
        prioritized - Prioritized replay buffer (Schaul et al., 2015)
      optimizer: str, name of optimizer to use.
      summary_writer: SummaryWriter object for outputting training statistics.
        Summary writing disabled if set to None.
      summary_writing_frequency: int, frequency with which summaries will be
        written. Lower values will result in slower training.
      seed: int, a seed for DQN's internal RNG, used for initialization and
        sampling actions. If None, will use the current time in nanoseconds.
      allow_partial_reload: bool, whether we allow reloading a partial agent
        (for instance, only the network parameters).
    """
    self._num_atoms = num_atoms
    self._kappa = kappa
    self._replay_scheme = replay_scheme

    super(JaxQuantileAgent, self).__init__(
        num_actions=num_actions,
        observation_shape=observation_shape,
        observation_dtype=observation_dtype,
        stack_size=stack_size,
        network=functools.partial(network, num_atoms=num_atoms),
        gamma=gamma,
        update_horizon=update_horizon,
        min_replay_history=min_replay_history,
        update_period=update_period,
        target_update_period=target_update_period,
        epsilon_fn=epsilon_fn,
        epsilon_train=epsilon_train,
        epsilon_eval=epsilon_eval,
        epsilon_decay_period=epsilon_decay_period,
        optimizer=optimizer,
        summary_writer=summary_writer,
        summary_writing_frequency=summary_writing_frequency,
        seed=seed,
        allow_partial_reload=allow_partial_reload)

  def _build_networks_and_optimizer(self):
    self._rng, rng = jax.random.split(self._rng)
    self.online_params = self.network_def.init(rng, x=self.state)
    self.optimizer = dqn_agent.create_optimizer(self._optimizer_name)
    self.optimizer_state = self.optimizer.init(self.online_params)
    self.target_network_params = self.online_params

  def _build_replay_buffer(self):
    """Creates the replay buffer used by the agent."""
    if self._replay_scheme not in ['uniform', 'prioritized']:
      raise ValueError('Invalid replay scheme: {}'.format(self._replay_scheme))
    # Both replay schemes use the same data structure, but the 'uniform' scheme
    # sets all priorities to the same value (which yields uniform sampling).
    return prioritized_replay_buffer.OutOfGraphPrioritizedReplayBuffer(
        observation_shape=self.observation_shape,
        stack_size=self.stack_size,
        update_horizon=self.update_horizon,
        gamma=self.gamma,
        observation_dtype=self.observation_dtype)

  def _train_step(self):
    """Runs a single training step.

    Runs training if both:
      (1) A minimum number of frames have been added to the replay buffer.
      (2) `training_steps` is a multiple of `update_period`.

    Also, syncs weights from online_params to target_network_params if training
    steps is a multiple of target update period.
    """
    if self._replay.add_count > self.min_replay_history:
      if self.training_steps % self.update_period == 0:
        self._sample_from_replay_buffer()
        self.optimizer_state, self.online_params, loss, mean_loss = train(
            self.network_def,
            self.online_params,
            self.target_network_params,
            self.optimizer,
            self.optimizer_state,
            self.preprocess_fn(self.replay_elements['state']),
            self.replay_elements['action'],
            self.preprocess_fn(self.replay_elements['next_state']),
            self.replay_elements['reward'],
            self.replay_elements['terminal'],
            self._kappa,
            self._num_atoms,
            self.cumulative_gamma)
        if self._replay_scheme == 'prioritized':
          # The original prioritized experience replay uses a linear exponent
          # schedule 0.4 -> 1.0. Comparing the schedule to a fixed exponent of
          # 0.5 on 5 games (Asterix, Pong, Q*Bert, Seaquest, Space Invaders)
          # suggested a fixed exponent actually performs better, except on Pong.
          probs = self.replay_elements['sampling_probabilities']
          loss_weights = 1.0 / jnp.sqrt(probs + 1e-10)
          loss_weights /= jnp.max(loss_weights)

          # Rainbow and prioritized replay are parametrized by an exponent
          # alpha, but in both cases it is set to 0.5 - for simplicity's sake we
          # leave it as is here, using the more direct sqrt(). Taking the square
          # root "makes sense", as we are dealing with a squared loss.  Add a
          # small nonzero value to the loss to avoid 0 priority items. While
          # technically this may be okay, setting all items to 0 priority will
          # cause troubles, and also result in 1.0 / 0.0 = NaN correction terms.
          self._replay.set_priority(self.replay_elements['indices'],
                                    jnp.sqrt(loss + 1e-10))

          # Weight the loss by the inverse priorities.
          loss = loss_weights * loss
          mean_loss = jnp.mean(loss)
        if self.summary_writer is not None:
          summary = tf.compat.v1.Summary(value=[
              tf.compat.v1.Summary.Value(tag='QuantileLoss',
                                         simple_value=mean_loss)])
          self.summary_writer.add_summary(summary, self.training_steps)
      if self.training_steps % self.target_update_period == 0:
        self._sync_weights()

    self.training_steps += 1



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Atari 100k rainbow agent with support for data augmentation."""

import functools
from absl import logging
from dopamine.jax.agents.full_rainbow import full_rainbow_agent
import gin
import jax
import jax.numpy as jnp
import tensorflow as tf


############################ Data Augmentation ############################
def _random_crop(key, img, cropped_shape):
  """Random crop an image."""
  _, width, height = cropped_shape[:-1]
  key_x, key_y = jax.random.split(key, 2)
  x = jax.random.randint(key_x, shape=(), minval=0, maxval=img.shape[1] - width)
  y = jax.random.randint(
      key_y, shape=(), minval=0, maxval=img.shape[2] - height)
  return img[:, x:x + width, y:y + height]


def _intensity_aug(key, x, scale=0.05):
  """Follows the code in Schwarzer et al. (2020) for intensity augmentation."""
  r = jax.random.normal(key, shape=(x.shape[0], 1, 1, 1))
  noise = 1.0 + (scale * jnp.clip(r, -2.0, 2.0))
  return x * noise


def drq_image_augmentation(key, obs, img_pad=4):
  """Padding and cropping for DrQ (Kostrikov et. al, 2020)."""
  paddings = [(0, 0), (img_pad, img_pad), (img_pad, img_pad), (0, 0)]
  cropped_shape = obs.shape
  # The reference uses ReplicationPad2d in pytorch, but it is not available
  # in Jax. Use 'edge' instead.
  obs = jnp.pad(obs, paddings, 'edge')
  key1, key2 = jax.random.split(key, num=2)
  aug_obs = _intensity_aug(key1, _random_crop(key2, obs, cropped_shape))
  return aug_obs


def preprocess_inputs_with_augmentation(x, data_augmentation=False, rng=None):
  """Input normalization and if specified, data augmentation."""
  out = x.astype(jnp.float32) / 255.
  if data_augmentation:
    if rng is None:
      raise ValueError('Pass rng when using data augmentation')
    out = drq_image_augmentation(rng, out)
  return out


@gin.configurable
class Atari100kRainbowAgent(full_rainbow_agent.JaxFullRainbowAgent):
  """A compact implementation of agents for Atari 100k."""

  def __init__(self,
               num_actions,
               data_augmentation=False,
               summary_writer=None,
               seed=None):
    """Creates the Rainbow-based agent for the Atari 100k benchmark.

    On Atari 100k, an agent is evaluated after 100k environment steps, which
    corresponds to 2-3 hours of game play, for training.
    Args:
      num_actions: int, number of actions the agent can take at any state.
      data_augmentation: bool, whether to use data augmentation.
      summary_writer: SummaryWriter object, for outputting training statistics.
      seed: int, a seed for Jax RNG and initialization.
    """
    super().__init__(
        num_actions=num_actions,
        preprocess_fn=preprocess_inputs_with_augmentation,
        summary_writer=summary_writer,
        seed=seed)
    logging.info('\t data_augmentation: %s', data_augmentation)
    self._data_augmentation = data_augmentation
    logging.info('\t data_augmentation: %s', data_augmentation)
    # Preprocessing during training and evaluation can be possibly different,
    # for example, when using data augmentation during training.
    self.train_preprocess_fn = functools.partial(
        preprocess_inputs_with_augmentation,
        data_augmentation=data_augmentation)

  def _training_step_update(self):
    """Gradient update during every training step."""

    self._sample_from_replay_buffer()
    self._rng, rng1, rng2 = jax.random.split(self._rng, num=3)
    states = self.train_preprocess_fn(self.replay_elements['state'], rng=rng1)
    next_states = self.train_preprocess_fn(
        self.replay_elements['next_state'], rng=rng2)

    if self._replay_scheme == 'prioritized':
      probs = self.replay_elements['sampling_probabilities']
      # Weight the loss by the inverse priorities.
      loss_weights = 1.0 / jnp.sqrt(probs + 1e-10)
      loss_weights /= jnp.max(loss_weights)
    else:
      # Uniform weights if not using prioritized replay.
      loss_weights = jnp.ones(states.shape[0])

    (self.optimizer_state, self.online_params,
     loss, mean_loss, self._rng) = full_rainbow_agent.train(
         self.network_def, self.online_params, self.target_network_params,
         self.optimizer, self.optimizer_state, states,
         self.replay_elements['action'], next_states,
         self.replay_elements['reward'], self.replay_elements['terminal'],
         loss_weights, self._support, self.cumulative_gamma, self._double_dqn,
         self._distributional, self._rng)

    if self._replay_scheme == 'prioritized':
      self._replay.set_priority(self.replay_elements['indices'],
                                jnp.sqrt(loss + 1e-10))

    if self.summary_writer is not None:
      summary = tf.compat.v1.Summary(value=[
          tf.compat.v1.Summary.Value(
              tag='CrossEntropyLoss', simple_value=mean_loss)
      ])
      self.summary_writer.add_summary(summary, self.training_steps)




# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MinAtar environment made compatible for Dopamine."""

import time

from dopamine.discrete_domains import atari_lib
from flax import linen as nn
import gin
import jax
import jax.numpy as jnp
import minatar

gin.constant('minatar_env.ASTERIX_SHAPE', (10, 10, 4))
gin.constant('minatar_env.BREAKOUT_SHAPE', (10, 10, 4))
gin.constant('minatar_env.FREEWAY_SHAPE', (10, 10, 7))
gin.constant('minatar_env.SEAQUEST_SHAPE', (10, 10, 10))
gin.constant('minatar_env.SPACE_INVADERS_SHAPE', (10, 10, 6))
gin.constant('minatar_env.DTYPE', jnp.float64)


class MinAtarEnv(object):
  """Wrapper class for MinAtar environments."""

  def __init__(self, game_name):
    self.env = minatar.Environment(env_name=game_name)
    self.env.n = self.env.num_actions()
    self.game_over = False

  @property
  def observation_space(self):
    return self.env.state_shape()

  @property
  def action_space(self):
    return self.env  # Only used for the `n` parameter.

  @property
  def reward_range(self):
    pass  # Unused

  @property
  def metadata(self):
    pass  # Unused

  def reset(self):
    self.game_over = False
    self.env.reset()
    return self.env.state()

  def step(self, action):
    r, terminal = self.env.act(action)
    self.game_over = terminal
    return self.env.state(), r, terminal, None


class StickyMinAtarEnv(MinAtarEnv):
  """Wrapper class for MinAtar environments with sticky actions."""

  def __init__(self, game_name, action_repeat_probability, seed=None):
    super().__init__(game_name)
    self.action_repeat_probability = action_repeat_probability
    self._prev_action = None
    seed = int(time.time() * 1e6) if seed is None else seed
    self._rng = jax.random.PRNGKey(seed)

  def reset(self):
    self._prev_action = None
    return super().reset()

  def step(self, action):
    action_to_use = action
    if self._prev_action:
      self._rng, rng = jax.random.split(self._rng, num=2)
      if jax.random.uniform(rng) < self.action_repeat_probability:
        action_to_use = self._prev_action
    self._prev_action = action
    return super().step(action_to_use)


@gin.configurable
def create_minatar_env(game_name,
                       sticky_actions=False,
                       action_repeat_probability=0.25,
                       seed=None):
  """Helper function for creating Minatar environment."""
  if sticky_actions:
    return StickyMinAtarEnv(
        game_name,
        action_repeat_probability=action_repeat_probability,
        seed=seed)
  return MinAtarEnv(game_name)


@gin.configurable
class MinatarDQNNetwork(nn.Module):
  """JAX DQN Network for Minatar environments."""
  num_actions: int
  inputs_preprocessed: bool = True

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
    initializer = nn.initializers.xavier_uniform()
    x = x.astype(jnp.float32)
    x = nn.Conv(features=16, kernel_size=(3, 3), strides=(1, 1),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = x.reshape(-1)  # flatten
    q_values = nn.Dense(features=self.num_actions, kernel_init=initializer)(x)
    return atari_lib.DQNNetworkType(q_values)


@gin.configurable
class MinatarRainbowNetwork(nn.Module):
  """Jax Rainbow network for Minatar."""
  num_actions: int
  num_atoms: int
  inputs_preprocessed: bool = True

  @nn.compact
  def __call__(self, x: jnp.ndarray, support: jnp.ndarray) -> jnp.ndarray:
    initializer = nn.initializers.xavier_uniform()
    x = x.astype(jnp.float32)
    x = nn.Conv(features=16, kernel_size=(3, 3), strides=(1, 1),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = x.reshape(-1)  # flatten
    x = nn.Dense(features=self.num_actions * self.num_atoms,
                 kernel_init=initializer)(x)
    logits = x.reshape((self.num_actions, self.num_atoms))
    probabilities = nn.softmax(logits)
    q_values = jnp.sum(support * probabilities, axis=1)
    return atari_lib.RainbowNetworkType(q_values, logits, probabilities)


@gin.configurable
class MinatarQuantileNetwork(nn.Module):
  """Convolutional network used to compute the agent's return quantiles."""
  num_actions: int
  num_atoms: int
  inputs_preprocessed: bool = True

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
    initializer = jax.nn.initializers.variance_scaling(
        scale=1.0 / jnp.sqrt(3.0),
        mode='fan_in',
        distribution='uniform')
    x = x.astype(jnp.float32)
    x = nn.Conv(features=16, kernel_size=(3, 3), strides=(1, 1),
                kernel_init=initializer)(x)
    x = nn.relu(x)
    x = x.reshape(-1)  # flatten
    x = nn.Dense(features=self.num_actions * self.num_atoms,
                 kernel_init=initializer)(x)
    logits = x.reshape((self.num_actions, self.num_atoms))
    probabilities = nn.softmax(logits)
    q_values = jnp.mean(logits, axis=1)
    return atari_lib.RainbowNetworkType(q_values, logits, probabilities)



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compact implementation of the full Rainbow agent in JAX.

Specifically, we implement the following components from Rainbow:

  * n-step updates
  * prioritized replay
  * distributional RL
  * double_dqn
  * noisy
  * dueling

Details in "Rainbow: Combining Improvements in Deep Reinforcement Learning" by
Hessel et al. (2018).
"""

import functools

from absl import logging
from dopamine.jax import losses
from dopamine.jax import networks
from dopamine.jax.agents.dqn import dqn_agent
from dopamine.jax.agents.rainbow import rainbow_agent
from dopamine.replay_memory import prioritized_replay_buffer
import gin
import jax
import jax.numpy as jnp
import numpy as onp
import optax
import tensorflow as tf


@functools.partial(jax.jit, static_argnums=(0, 4, 5, 6, 7, 8, 10, 11))
def select_action(network_def, params, state, rng, num_actions, eval_mode,
                  epsilon_eval, epsilon_train, epsilon_decay_period,
                  training_steps, min_replay_history, epsilon_fn, support):
  """Select an action from the set of available actions."""
  epsilon = jnp.where(
      eval_mode, epsilon_eval,
      epsilon_fn(epsilon_decay_period, training_steps, min_replay_history,
                 epsilon_train))

  rng, rng1, rng2, rng3 = jax.random.split(rng, num=4)
  p = jax.random.uniform(rng1)
  best_actions = jnp.argmax(
      network_def.apply(params, state, key=rng2, eval_mode=eval_mode,
                        support=support).q_values)
  return rng, jnp.where(p <= epsilon,
                        jax.random.randint(rng3, (), 0, num_actions),
                        best_actions)


@functools.partial(jax.vmap, in_axes=(None, 0, None))
def get_logits(model, states, rng):
  return model(states, key=rng).logits


@functools.partial(jax.vmap, in_axes=(None, 0, None))
def get_q_values(model, states, rng):
  return model(states, key=rng).q_values


@functools.partial(jax.jit, static_argnums=(0, 3, 12, 13, 14))
def train(network_def, online_params, target_params, optimizer, optimizer_state,
          states, actions, next_states, rewards, terminals, loss_weights,
          support, cumulative_gamma, double_dqn, distributional, rng):
  """Run a training step."""

  # Split the current rng into 2 for updating the rng after this call
  rng, rng1, rng2 = jax.random.split(rng, num=3)

  def q_online(state, key):
    return network_def.apply(online_params, state, key=key, support=support)

  def q_target(state, key):
    return network_def.apply(target_params, state, key=key, support=support)

  def loss_fn(params, target, loss_multipliers):
    """Computes the distributional loss for C51 or huber loss for DQN."""
    def q_online(state, key):
      return network_def.apply(params, state, key=key, support=support)

    if distributional:
      logits = get_logits(q_online, states, rng)
      logits = jnp.squeeze(logits)
      # Fetch the logits for its selected action. We use vmap to perform this
      # indexing across the batch.
      chosen_action_logits = jax.vmap(lambda x, y: x[y])(logits, actions)
      loss = jax.vmap(losses.softmax_cross_entropy_loss_with_logits)(
          target, chosen_action_logits)
    else:
      q_values = get_q_values(q_online, states, rng)
      q_values = jnp.squeeze(q_values)
      replay_chosen_q = jax.vmap(lambda x, y: x[y])(q_values, actions)
      loss = jax.vmap(losses.huber_loss)(target, replay_chosen_q)

    mean_loss = jnp.mean(loss_multipliers * loss)
    return mean_loss, loss

  # Use the weighted mean loss for gradient computation.
  grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
  target = target_output(q_online, q_target, next_states, rewards,
                         terminals, support, cumulative_gamma, double_dqn,
                         distributional, rng1)

  # Get the unweighted loss without taking its mean for updating priorities.
  (mean_loss, loss), grad = grad_fn(online_params, target, loss_weights)
  updates, optimizer_state = optimizer.update(grad, optimizer_state,
                                              params=online_params)
  online_params = optax.apply_updates(online_params, updates)
  return optimizer_state, online_params, loss, mean_loss, rng2


@functools.partial(
    jax.vmap, in_axes=(None, None, 0, 0, 0, None, None, None, None, None))
def target_output(model, target_network, next_states, rewards, terminals,
                  support, cumulative_gamma, double_dqn, distributional, rng):
  """Builds the C51 target distribution or DQN target Q-values."""

  is_terminal_multiplier = 1. - terminals.astype(jnp.float32)
  # Incorporate terminal state to discount factor.
  gamma_with_terminal = cumulative_gamma * is_terminal_multiplier

  target_network_dist = target_network(next_states, key=rng)
  if double_dqn:
    # Use the current network for the action selection
    next_state_target_outputs = model(next_states, key=rng)
  else:
    next_state_target_outputs = target_network_dist
  # Action selection using Q-values for next-state
  q_values = jnp.squeeze(next_state_target_outputs.q_values)
  next_qt_argmax = jnp.argmax(q_values)

  if distributional:
    # Compute the target Q-value distribution
    probabilities = jnp.squeeze(target_network_dist.probabilities)
    next_probabilities = probabilities[next_qt_argmax]
    target_support = rewards + gamma_with_terminal * support
    target = rainbow_agent.project_distribution(
        target_support, next_probabilities, support)
  else:
    # Compute the target Q-value
    next_q_values = jnp.squeeze(target_network_dist.q_values)
    replay_next_qt_max = next_q_values[next_qt_argmax]
    target = rewards + gamma_with_terminal * replay_next_qt_max

  return jax.lax.stop_gradient(target)


@gin.configurable
class JaxFullRainbowAgent(dqn_agent.JaxDQNAgent):
  """A compact implementation of the full Rainbow agent."""

  def __init__(self,
               num_actions,
               noisy=True,
               dueling=True,
               double_dqn=True,
               distributional=True,
               num_updates_per_train_step=1,
               network=networks.FullRainbowNetwork,
               num_atoms=51,
               vmax=10.,
               vmin=None,
               epsilon_fn=dqn_agent.linearly_decaying_epsilon,
               replay_scheme='prioritized',
               summary_writer=None,
               seed=None,
               preprocess_fn=None):
    """Initializes the agent and constructs the necessary components.

    Args:
      num_actions: int, number of actions the agent can take at any state.
      noisy: bool, Whether to use noisy networks or not.
      dueling: bool, Whether to use dueling network architecture or not.
      double_dqn: bool, Whether to use Double DQN or not.
      distributional: bool, whether to use distributional RL or not.
      num_updates_per_train_step: int, Number of gradient updates every training
        step. Defaults to 1.
      network: flax.linen Module, neural network used by the agent initialized
        by shape in _create_network below. See
        dopamine.jax.networks.RainbowNetwork as an example.
      num_atoms: int, the number of buckets of the value function distribution.
      vmax: float, the value distribution support is [vmin, vmax].
      vmin: float, the value distribution support is [vmin, vmax]. If vmin is
        None, it is set to -vmax.
      epsilon_fn: function expecting 4 parameters: (decay_period, step,
        warmup_steps, epsilon). This function should return the epsilon value
        used for exploration during training.
      replay_scheme: str, 'prioritized' or 'uniform', the sampling scheme of the
        replay memory.
      summary_writer: SummaryWriter object, for outputting training statistics.
      seed: int, a seed for Jax RNG and initialization.
      preprocess_fn: function expecting the input state as parameter which
        it preprocesses (such as normalizing the pixel values between 0 and 1)
        before passing it to the Q-network. Defaults to None.
    """
    logging.info('Creating %s agent with the following parameters:',
                 self.__class__.__name__)
    logging.info('\t double_dqn: %s', double_dqn)
    logging.info('\t noisy_networks: %s', noisy)
    logging.info('\t dueling_dqn: %s', dueling)
    logging.info('\t distributional: %s', distributional)
    logging.info('\t num_atoms: %d', num_atoms)
    logging.info('\t replay_scheme: %s', replay_scheme)
    logging.info('\t num_updates_per_train_step: %d',
                 num_updates_per_train_step)
    # We need this because some tools convert round floats into ints.
    vmax = float(vmax)
    self._num_atoms = num_atoms
    vmin = vmin if vmin else -vmax
    self._support = jnp.linspace(vmin, vmax, num_atoms)
    self._replay_scheme = replay_scheme
    self._double_dqn = double_dqn
    self._noisy = noisy
    self._dueling = dueling
    self._distributional = distributional
    self._num_updates_per_train_step = num_updates_per_train_step

    super().__init__(
        num_actions=num_actions,
        network=functools.partial(
            network, num_atoms=num_atoms,
            noisy=self._noisy,
            dueling=self._dueling,
            distributional=self._distributional),
        epsilon_fn=dqn_agent.identity_epsilon if self._noisy else epsilon_fn,
        summary_writer=summary_writer,
        seed=seed,
        preprocess_fn=preprocess_fn)

  def _build_networks_and_optimizer(self):
    self._rng, rng = jax.random.split(self._rng)
    self.online_params = self.network_def.init(rng, x=self.state,
                                               support=self._support)
    self.optimizer = dqn_agent.create_optimizer(self._optimizer_name)
    self.optimizer_state = self.optimizer.init(self.online_params)
    self.target_network_params = self.online_params

  def _build_replay_buffer(self):
    """Creates the replay buffer used by the agent."""
    if self._replay_scheme not in ['uniform', 'prioritized']:
      raise ValueError('Invalid replay scheme: {}'.format(self._replay_scheme))
    return prioritized_replay_buffer.OutOfGraphPrioritizedReplayBuffer(
        observation_shape=self.observation_shape,
        stack_size=self.stack_size,
        update_horizon=self.update_horizon,
        gamma=self.gamma,
        observation_dtype=self.observation_dtype)

  def _training_step_update(self):
    """Gradient update during every training step."""

    self._sample_from_replay_buffer()
    states = self.preprocess_fn(self.replay_elements['state'])
    next_states = self.preprocess_fn(self.replay_elements['next_state'])

    if self._replay_scheme == 'prioritized':
      # The original prioritized experience replay uses a linear exponent
      # schedule 0.4 -> 1.0. Comparing the schedule to a fixed exponent of
      # 0.5 on 5 games (Asterix, Pong, Q*Bert, Seaquest, Space Invaders)
      # suggested a fixed exponent actually performs better, except on Pong.
      probs = self.replay_elements['sampling_probabilities']
      # Weight the loss by the inverse priorities.
      loss_weights = 1.0 / jnp.sqrt(probs + 1e-10)
      loss_weights /= jnp.max(loss_weights)
    else:
      # Uniform weights if not using prioritized replay.
      loss_weights = jnp.ones(states.shape[0])

    (self.optimizer_state, self.online_params,
     loss, mean_loss, self._rng) = train(
         self.network_def, self.online_params, self.target_network_params,
         self.optimizer, self.optimizer_state, states,
         self.replay_elements['action'], next_states,
         self.replay_elements['reward'], self.replay_elements['terminal'],
         loss_weights, self._support, self.cumulative_gamma, self._double_dqn,
         self._distributional, self._rng)

    if self._replay_scheme == 'prioritized':
      # Rainbow and prioritized replay are parametrized by an exponent
      # alpha, but in both cases it is set to 0.5 - for simplicity's sake we
      # leave it as is here, using the more direct sqrt(). Taking the square
      # root "makes sense", as we are dealing with a squared loss.  Add a
      # small nonzero value to the loss to avoid 0 priority items. While
      # technically this may be okay, setting all items to 0 priority will
      # cause troubles, and also result in 1.0 / 0.0 = NaN correction terms.
      self._replay.set_priority(self.replay_elements['indices'],
                                jnp.sqrt(loss + 1e-10))

    if self.summary_writer is not None:
      summary = tf.compat.v1.Summary(value=[
          tf.compat.v1.Summary.Value(
              tag='CrossEntropyLoss', simple_value=mean_loss)
      ])
      self.summary_writer.add_summary(summary, self.training_steps)

  def _store_transition(self,
                        last_observation,
                        action,
                        reward,
                        is_terminal,
                        *args,
                        priority=None,
                        episode_end=False):
    """Stores a transition when in training mode."""
    is_prioritized = isinstance(
        self._replay,
        prioritized_replay_buffer.OutOfGraphPrioritizedReplayBuffer)
    if is_prioritized and priority is None:
      if self._replay_scheme == 'uniform':
        priority = 1.
      else:
        priority = self._replay.sum_tree.max_recorded_priority

    if not self.eval_mode:
      self._replay.add(
          last_observation,
          action,
          reward,
          is_terminal,
          *args,
          priority=priority,
          episode_end=episode_end)

  def _train_step(self):
    """Runs a single training step.

    Runs training if both:
      (1) A minimum number of frames have been added to the replay buffer.
      (2) `training_steps` is a multiple of `update_period`.

    Also, syncs weights from online_network_params to target_network_params if
    training steps is a multiple of target update period.
    """
    if self._replay.add_count > self.min_replay_history:
      if self.training_steps % self.update_period == 0:
        for _ in range(self._num_updates_per_train_step):
          self._training_step_update()

      if self.training_steps % self.target_update_period == 0:
        self._sync_weights()

    self.training_steps += 1

  def begin_episode(self, observation):
    """Returns the agent's first action for this episode."""
    self._reset_state()
    self._record_observation(observation)

    if not self.eval_mode:
      self._train_step()

    state = self.preprocess_fn(self.state)
    self._rng, self.action = select_action(
        self.network_def, self.online_params, state, self._rng,
        self.num_actions, self.eval_mode, self.epsilon_eval, self.epsilon_train,
        self.epsilon_decay_period, self.training_steps, self.min_replay_history,
        self.epsilon_fn, self._support)
    self.action = onp.asarray(self.action)
    return self.action

  def step(self, reward, observation):
    """Records the most recent transition and returns the agent's next action."""
    self._last_observation = self._observation
    self._record_observation(observation)

    if not self.eval_mode:
      self._store_transition(self._last_observation, self.action, reward, False)
      self._train_step()

    state = self.preprocess_fn(self.state)
    self._rng, self.action = select_action(
        self.network_def, self.online_params, state, self._rng,
        self.num_actions, self.eval_mode, self.epsilon_eval, self.epsilon_train,
        self.epsilon_decay_period, self.training_steps, self.min_replay_history,
        self.epsilon_fn, self._support)
    self.action = onp.asarray(self.action)
    return self.action



# coding=utf-8
# Copyright 2021 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Deepmind Control Suite utilities.

Includes a wrapper class around Deepmind Control Suite environments.
This class makes general Deepmind Control Suite environments conformant with
the API Dopamine is expecting.
"""

import collections
from typing import Any, Mapping, Tuple

from dm_control import suite
from dm_control.rl import control
import dm_env
from dm_env import specs
import gin
import gym
from gym import spaces
import numpy as np


@gin.configurable
def create_deepmind_control_environment(
    domain_name: str = gin.REQUIRED,
    task_name: str = gin.REQUIRED,
    use_image_observations: bool = gin.REQUIRED) -> gym.Env:
  """Wraps a Deepmind Control Suite environment with some basic preprocessing.

  Args:
    domain_name: The name of the environment to run.
    task_name: The name of the task to run.
    use_image_observations: If True, the created environment will return image
      observations. Else, it will return the state vector.

  Returns:
    A Gym environment with some standard preprocessing.
  """
  env = suite.load(
      domain_name, task_name, environment_kwargs={'flat_observation': True})

  # Wrap the returned environment in a class which conforms to the API expected
  # by Dopamine.
  if use_image_observations:
    env = DeepmindControlWithImagesPreprocessing(env)
  else:
    env = DeepmindControlPreprocessing(env)

  return env


@gin.configurable(allowlist=['action_repeat'])
class DeepmindControlPreprocessing(gym.Env):
  """A DM Control Suite preprocessing wrapper.

  Attributes:
    environment: The underlying environment that is wrapped for preprocessing.
    game_over: True when the last environment step led to a game over signal.
  """

  def __init__(self, environment: control.Environment, action_repeat: int = 1):
    """Initializes a preprocessing wrapper for a deepmind control environment.

    Args:
      environment: The environment to wrap for preprocessing.
      action_repeat: The number of times to repeat a supplied action. Must be
        greater than 0.
    """
    if action_repeat < 1:
      raise ValueError(f'Action repeat must be > 0. Got {action_repeat}.')

    self.environment = environment
    self._action_repeat = action_repeat
    self.game_over = False

  @property
  def observation_space(self) -> spaces.Box:
    """The observation space of the processed environment."""
    obs_spec = self.environment.observation_spec()
    assert isinstance(obs_spec, collections.OrderedDict)
    # Since observations are flattened, all shapes are 1 dimensional
    low = list()
    high = list()
    for k in obs_spec:
      assert isinstance(obs_spec[k], specs.Array)
      if isinstance(obs_spec[k], specs.BoundedArray):
        low.extend(obs_spec[k].minimum.tolist())
        high.extend(obs_spec[k].maximum.tolist())
      else:
        low.extend([-float('inf') for _ in range(obs_spec[k].shape[0])])
        high.extend([float('inf') for _ in range(obs_spec[k].shape[0])])
    low = np.asarray(low, dtype=np.float32)
    high = np.asarray(high, dtype=np.float32)
    return spaces.Box(low=low, high=high, dtype=np.float32)

  @property
  def action_space(self) -> spaces.Box:
    """The action space for the processed environment."""
    action_spec = self.environment.action_spec()
    return spaces.Box(
        low=action_spec.minimum, high=action_spec.maximum, dtype=np.float32)

  @property
  def reward_range(self) -> Tuple[float, float]:
    """The reward range for the processed environment."""
    return (-float('inf'), float('inf'))

  @property
  def metadata(self) -> Mapping[Any, Any]:
    """The metadata for the processed environment."""
    return {}

  def reset(self) -> np.ndarray:
    """Resets the environment.

    Returns:
      An initial observation.
    """
    timestep = self.environment.reset()
    return self._get_observation(timestep)

  def step(
      self,
      action: np.ndarray) -> Tuple[np.ndarray, float, bool, Mapping[Any, Any]]:
    """Steps the environment.

    Args:
      action: The action to apply.

    Returns:
      An (observation, reward, is_terminal, info) tuple. For more information,
      see base class.
    """
    accumulated_reward = 0.0
    self.game_over = False
    done = False

    for _ in range(self._action_repeat):
      timestep = self.environment.step(action)
      accumulated_reward += timestep.reward

      if timestep.last():
        self.game_over = True  # Signals end of episode.

        # Only set as terminal if not the result of a timeout.
        done = timestep.discount < 1.0
        break

    observation = self._get_observation(timestep)

    return observation, accumulated_reward, done, {}

  def _get_observation(self, timestep: dm_env.TimeStep) -> np.ndarray:
    return np.concatenate(
        [timestep.observation[k] for k in timestep.observation])


class DeepmindControlWithImagesPreprocessing(DeepmindControlPreprocessing):
  """A DM Control Suite preprocessing wrapper for image observations."""

  def __init__(self,
               env: control.Environment,
               observation_shape: Tuple[int, int] = (84, 84)):
    """Constructor for preprocessing wrapper.

    Args:
      env: The environment to wrap.
      observation_shape: The size to reshape the images to. This corresponds
        to (height, width). The output shape will be (height, width, 3), with
        3 corresponding to RGB channels.
    """
    super(DeepmindControlWithImagesPreprocessing, self).__init__(env)

    self._shape = observation_shape

  @property
  def observation_space(self) -> spaces.Box:
    return spaces.Box(low=0, high=255, shape=(*self._shape, 3,), dtype=np.uint8)

  def _get_observation(self, timestep: dm_env.TimeStep) -> np.ndarray:
    return self._render_image()

  def _render_image(self) -> np.ndarray:
    """Renders the environment and processes the image.

    The processing steps are:
    1. Convert the image to uint8, as expected by Dopamine's replay buffer.
    2. Resize the image to the specified size.

    Returns:
      An np.ndarray with the image data in rgb format.
    """
    image = self.environment.physics.render(*self._shape, camera_id=0)

    if image.dtype != np.uint8:
      if image.dtype in (np.float32, np.float64):
        image = (image * 255).astype(np.uint8)
      else:
        raise ValueError('Unsupported dtype: {}'.format(image.dtype))

    return image
